"""Convert colours to sepia."""
from dataclasses import dataclass

import numpy as np

from .base_matrix import BaseMatrix


SRC_ARR = np.array([
    [0.393, 0.769, 0.189],
    [0.349, 0.686, 0.168],
    [0.272, 0.534, 0.131],
])

DEST_ARR = np.array([
    [0.607, -0.769, -0.189],
    [-0.349, 0.314, -0.168],
    [-0.272, -0.534, 0.869],
])


@dataclass
class Sepia(BaseMatrix):
    """A sepia effect."""
    amount: float = 1.0

    def get_matrix(self):
        matrix = np.identity(4, dtype=np.float32)
        matrix[:3, :3] = (1 - self.amount) * DEST_ARR
        matrix[:3, :3] += SRC_ARR
        return matrix



"""Manage the compilation of programs in a context."""
from typing import Optional, Dict, Tuple
from contextlib import contextmanager
import pkgutil

import moderngl
import numpy as np


class ShaderManager:
    """Cache of compiled programs for a context.

    The shadermanager is cached onto the context object.

    """
    ctx: moderngl.Context
    files: Dict[str, str]
    programs: Dict[Tuple[str, str, Optional[str]], moderngl.Program]

    def __init__(self, ctx: moderngl.Context):
        self.ctx = ctx
        if not ctx.extra:
            ctx.extra = {}
        if 'shadermgr' in ctx.extra:
            raise ValueError(f"ShaderManager is already defined for {ctx}")
        ctx.extra['shadermgr'] = self
        self.files = {}
        self.programs = {}

    def _read(self, name: str):
        """Read a GLSL file from the wasabi2d directory."""
        return pkgutil.get_data('wasabi2d', 'glsl/' + name).decode('utf-8')

    def load(self, *names) -> moderngl.Program:
        """Load a program from the wasabi2d/glsl directory.

        If a single argument is given, load shaders from files with a common
        name, and use a geometry shader if it exists.

        Also accepts two or three arguments to specify separate filenames for
        each shader stage. Stages are in vert, geom, frag order and the geom
        stage is optional.

        """
        if names in self.programs:
            return self.programs[names]

        if len(names) == 1:
            name = names[0]
            try:
                geom_shader = self._read(f'{name}.geom')
            except FileNotFoundError:
                geom_shader = None
            prog = self.programs[names] = self.ctx.program(
                vertex_shader=self._read(f'{name}.vert'),
                geometry_shader=geom_shader,
                fragment_shader=self._read(f'{name}.frag'),
            )
        elif len(names) == 2:
            vert, frag = names
            prog = self.programs[names] = self.ctx.program(
                vertex_shader=self._read(f'{vert}.vert'),
                fragment_shader=self._read(f'{frag}.frag'),
            )
        elif len(names) == 3:
            vert, geom, frag = names
            prog = self.programs[names] = self.ctx.program(
                vertex_shader=self._read(f'{vert}.vert'),
                geometry_shader=self._read(f'{geom}.geom'),
                fragment_shader=self._read(f'{frag}.frag'),
            )
        else:
            raise ValueError("Unexpected number of files.")
        return prog

    def get(
        self,
        vertex_shader: str,
        fragment_shader: str,
        geometry_shader: Optional[str] = None
    ) -> moderngl.Program:
        """Get a compiled program."""
        assert isinstance(vertex_shader, str)
        assert isinstance(fragment_shader, str)
        assert isinstance(geometry_shader, (str, type(None)))
        k = vertex_shader, fragment_shader, geometry_shader
        try:
            return self.programs[k]
        except KeyError:
            pass

        prog = self.programs[k] = self.ctx.program(
            vertex_shader=vertex_shader,
            fragment_shader=fragment_shader,
            geometry_shader=geometry_shader,
        )
        return prog

    def set_proj(self, proj: np.ndarray):
        """Set the projection matrix."""
        projmatrix = proj.tobytes()
        for prog in self.programs.values():
            try:
                uniform = prog['proj']
            except KeyError:
                continue
            uniform.write(projmatrix)

    def __del__(self):
        self.release()

    def release(self):
        for prog in self.programs.values():
            prog.release()
        self.programs.clear()


def shadermgr(ctx: moderngl.Context) -> ShaderManager:
    """Shortcut to get or create the shader manager for a context."""
    return ctx.extra['shadermgr']


@contextmanager
def bind_framebuffer(ctx, fb, *, clear=False):
    """Bind an alternative framebuffer during a context.

    The previous binding will be restored when the context exits.

    If `clear` is True, also clear the framebuffer.

    """
    orig_screen = ctx._screen
    ctx._screen = fb
    try:
        fb.use()
        if clear:
            fb.clear(0, 0, 0, 0)
        yield
    finally:
        orig_screen.use()
        ctx._screen = orig_screen


blend_aliases = {
    '1': moderngl.ONE,
    1: moderngl.ONE,
    '0': moderngl.ZERO,
    0: moderngl.ZERO,
    'a': moderngl.SRC_ALPHA,
    'da': moderngl.DST_ALPHA,
    '1-a': moderngl.ONE_MINUS_SRC_ALPHA,
    '1-da': moderngl.ONE_MINUS_DST_ALPHA,
}


@contextmanager
def blend_func(
    ctx,
    src=moderngl.SRC_ALPHA,
    dest=moderngl.ONE_MINUS_SRC_ALPHA,
    src_a=moderngl.ONE,
    dest_a=moderngl.ONE_MINUS_SRC_ALPHA,
):
    """Override the blending function for the duration of the context."""
    src = blend_aliases.get(src, src)
    dest = blend_aliases.get(dest, dest)
    src_a = blend_aliases.get(src_a, src_a)
    dest_a = blend_aliases.get(dest_a, dest_a)

    ctx.blend_func = src, dest, src_a, dest_a
    try:
        yield
    finally:
        ctx.blend_func = (
            moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
            moderngl.ONE, moderngl.ONE_MINUS_SRC_ALPHA
        )



"""Example of using one layer to mask another."""
import wasabi2d as w2d

scene = w2d.Scene()
center = (scene.width / 2, scene.height / 2)

photo = scene.layers[0].add_sprite(
    'positano',
    pos=center
)
photo.scale = max(
    scene.width / photo.width,
    scene.height / photo.height
)

mask = scene.layers[1].add_circle(
    radius=100,
    pos=center,
)

scene.chain = [
    w2d.chain.Mask(
        w2d.chain.Layers([1]),
        w2d.chain.LayerRange(stop=0),
    )
]


@w2d.event
def on_mouse_move(pos):
    mask.pos = pos


@w2d.event
def on_mouse_down(pos):
    w2d.animate(mask, tween='decelerate', duration=0.2, scale=2)


@w2d.event
def on_mouse_up(pos):
    w2d.animate(mask, tween='bounce_end', duration=0.5, scale=1)


w2d.run()



from dataclasses import dataclass

import numpy as np
import moderngl

from ..descriptors import CallbackProp
from .base import Colorable, Transformable, Bounds
from ..allocators.packed import PackedBuffer


QUAD = np.array([0, 1, 2, 0, 2, 3], dtype='u4')


@dataclass
class TextureContext:
    tex: moderngl.Texture
    prog: moderngl.Program

    def __enter__(self):
        """Bind the given texture to the given program during the context."""
        self.prog['tex'].value = 0
        self.tex.use(0)

    def __exit__(self, *_):
        pass


class Sprite(Colorable, Transformable):
    """A sprite is a quad with an image texture."""

    def __init__(
            self,
            layer,
            image,
            anchor_x='center',
            anchor_y='center'):
        super().__init__()
        self.verts = None
        self.layer = layer
        self._image = None
        self._anchor_x = anchor_x
        self._anchor_y = anchor_y
        self._vert_color = np.ones((4, 4), dtype='f4')

        self._array = None
        self._array_id = None

        self.image = image  # trigger sprite load and migration

    @property
    def image(self):
        """Get the name of the image for this sprite."""
        return self._image

    @image.setter
    def image(self, name):
        """Set the image."""
        if name == self._image:
            return

        self._image = name
        texregion = self.texregion = self.layer.group.atlas.get(name)

        self.uvs = texregion.texcoords
        self.orig_verts = texregion.get_verts(self._anchor_x, self._anchor_y)
        xs = self.orig_verts[:, 0]
        ys = self.orig_verts[:, 1]
        self.width = np.fabs(np.min(xs) - np.max(xs))
        self.height = np.fabs(np.min(ys) - np.max(ys))

        self._dirty = True

        tex = self.texregion.tex

        if not self._array:
            # migrate into a new array
            self._array = self._get_array(tex)
            self._array_id, _ = self._array.alloc(4, QUAD)
        elif tex is not self._array.draw_context.tex:
            # migrate out of this buffer
            self._array.remove(self._array_id)
            self._array = self._get_array(tex)
            self._array_id, _ = self._array.alloc(4, QUAD)

    def _get_array(self, tex):
        k = ('sprite', id(tex))
        array = self.layer.arrays.get(k)
        if not array:
            prog = self.layer.group.shadermgr.load('texquads')
            array = PackedBuffer(
                moderngl.TRIANGLES,
                self.layer.ctx,
                prog,
                dtype=np.dtype([
                    ('in_vert', '2f4'),
                    ('in_color', '4f2'),
                    ('in_uv', '2u2'),
                ]),
                draw_context=TextureContext(tex, prog),
            )
            self.layer.arrays[k] = array
        return array

    def _reset_verts(self):
        self.orig_verts = None
        self._set_dirty()

    anchor_x = CallbackProp(_reset_verts)
    anchor_y = CallbackProp(_reset_verts)

    def delete(self):
        """Delete this sprite."""
        self.layer._dirty.discard(self)
        self.layer.objects.discard(self)
        self._array.remove(self._array_id)
        self.layer = None
        self._array = self._array_id = None

    def _set_dirty(self):
        if self.layer:
            self.layer._dirty.add(self)
        self.verts = None

    bounds = Bounds('self.orig_verts[:, :2]')

    def _update(self):
        if self.orig_verts is None:
            self.orig_verts = self.texregion.get_verts(
                self._anchor_x,
                self._anchor_y
            )

        xform = self._xform()

        verts = self._array.get_verts(self._array_id)
        verts['in_color'][:] = self._color
        verts['in_uv'][:] = self.texregion.texcoords

        np.matmul(
            self.orig_verts,
            xform[:, :2],
            out=verts['in_vert']
        )



"""Example of the punch effect."""
import math
from wasabi2d import run, Scene, event


scene = Scene()

logo1 = scene.layers[0].add_sprite(
    'wasabi2d',
    pos=(scene.width / 4, scene.height / 2),
    scale=0.5
)
effect1 = scene.layers[0].set_effect(
    'pixellate',
    pxsize=10
)

logo2 = scene.layers[1].add_sprite(
    'wasabi2d',
    pos=(scene.width * 3 / 4, scene.height / 2),
    scale=0.5
)
effect2 = scene.layers[1].set_effect(
    'pixellate',
    pxsize=10,
    antialias=0.0
)

scene.layers[2].add_label(
    f'antialias={effect1.antialias}',
    pos=(scene.width / 4, scene.height - 20),
    align='center',
)
scene.layers[2].add_label(
    f'antialias={effect2.antialias}',
    pos=(scene.width * 3 / 4, scene.height - 20),
    align='center',
)


@event
def update(dt, t):
    logo1.angle += 0.5 * dt
    logo2.angle += 0.5 * dt


run()



"""Sparse Vertex buffer with a packed index buffer."""
from typing import Dict, Tuple, ContextManager
from contextlib import nullcontext

import moderngl
import numpy as np

from .index import IndexBuffer
from .vertlists import dtype_to_moderngl, MemoryBackedBuffer


class PackedBuffer:
    def __init__(
            self,
            mode: int,
            ctx: moderngl.Context,
            prog: moderngl.Program,
            dtype: np.dtype,
            draw_context: ContextManager = nullcontext(),
            capacity: int = 256,
            index_capacity: int = 512):
        self.mode = mode
        self.ctx = ctx
        self.prog = prog
        self.dtype = dtype_to_moderngl(dtype)
        self.allocs: Dict[int, Tuple[slice, np.ndarray]] = {}
        self.verts = MemoryBackedBuffer(ctx, capacity, dtype)
        self.indexes = IndexBuffer(ctx)
        self.draw_context = draw_context
        self.dirty = False

    def empty(self) -> bool:
        """Return True if there are no allocations in this buffer."""
        return bool(self.allocs)

    def alloc(
        self,
        num_verts: int,
        indexes: np.ndarray
    ) -> Tuple[int, np.ndarray]:
        """Allocate a list from within this buffer.

        Return the allocated ID and an array view that can be used to set
        the vertex data immediately.

        The array view is not guaranteed to be valid after other array
        update operations.

        """
        vertoff, vertbuf = self.verts.allocate(num_verts)
        id = self.indexes.insert(indexes + vertoff.start)

        self.allocs[id] = vertoff
        self.dirty = True
        return id, vertbuf

    def insert(self, verts: np.ndarray, indexes: np.ndarray) -> int:
        """Allocate a list from within this buffer."""
        id, vertbuf = self.alloc(len(verts), indexes)
        vertbuf[:] = verts
        return id

    def realloc(self, id: int, verts: np.ndarray, indexes: np.ndarray):
        """Update an allocation."""
        vertoff, _ = self.allocs.pop(id)

        vertoff, vertbuf = self.verts.realloc(
            vertoff,
            len(verts),
        )
        vertbuf[:] = verts
        self.indexes.set_indexes(id, indexes + vertoff.start)
        self.allocs[id] = vertoff
        self.dirty = True

    def get_verts(self, id: int) -> np.ndarray:
        """Get the vertex slice for the given allocation.

        It is assumed that the vertices will be modified and need to be
        synced to the GL.

        It is not guaranteed that the slice will remain valid after other
        array update operations.

        """
        self.dirty = True
        vertoff = self.allocs[id]
        return self.verts.array[vertoff]

    def remove(self, id: int):
        """Remove a list from the array."""
        vertoff = self.allocs.pop(id)
        self.verts.free(vertoff)
        self.indexes.remove(id)
        self.dirty = True

    def get_vao(self):
        # TODO: use the dirty list to more accurately indicate which parts of
        # a buffer need updating.
        vbo = self.verts.get_buffer(self.dirty)
        ibo = self.indexes.get_buffer()

        # TODO: only recreate the VAO if buffers have changed
        vao = self.ctx.vertex_array(
            self.prog,
            [
                (vbo, *self.dtype),
            ],
            ibo
        )
        return vao

    def render(self, camera):
        """Render all lists."""
        if not self.allocs:
            return
        vao = self.get_vao()
        with self.draw_context:
            vao.render(self.mode)
        vao.release()

    def release(self):
        """Release this array."""
        self.verts.release()
        self.indexes.release()



import math
import numpy as np
import wasabi2d as w2d
from wasabi2d import Vector2, keys
from wasabi2d.keyboard import keyboard
from pygame import joystick

INVISIBLE = (0, 0, 0, 0)
GREEN = (0.3, 1.3, 0.3)
TRANSPARENT_GREEN = (*GREEN[:3], 0)
LINE_W = 1
SHIP_PTS = [(10, 0), (-5, 5), (-3, 0), (-5, -5)]

GRAVITY = 0.5e7
ACCEL = 100
ROTATION_SPEED = 2


def forward(ship, length=1) -> Vector2:
    """Get a vector in the direction of the ship."""
    v = Vector2()
    v.from_polar((length, math.degrees(ship.angle)))
    return v


def make_player(pos, angle=0):
    ship = scene.layers[0].add_polygon(
        SHIP_PTS,
        fill=False,
        color=GREEN,
        stroke_width=LINE_W,
    )
    ship.pos = ship.initial_pos = pos
    ship.angle = ship.initial_angle = angle
    ship.v = forward(ship, 160)
    ship.initial_v = Vector2(ship.v)
    ship.radius = 7
    ship.dead = False
    return ship

mode_720p = 1280, 720
mode_1080p = 1920, 1080
scene = w2d.Scene(*mode_720p, fullscreen=True)
scene.chain = [
    w2d.LayerRange()
    .wrap_effect('trails', alpha=0.4, fade=0.08)
    .wrap_effect('bloom', radius=8)
]

center = Vector2(scene.width, scene.height) * 0.5

score1 = scene.layers[0].add_label('0', pos=(10, 40), fontsize=30, color=GREEN)
score2 = scene.layers[0].add_label(
    '0',
    pos=(scene.width - 10, 40),
    align='right',
    fontsize=30,
    color=GREEN
)
score1.value = score2.value = 0

fps = scene.layers[0].add_label(
    'FPS: 60',
    pos=(10, scene.height - 10),
    fontsize=20,
    color=GREEN,
)

particles = scene.layers[0].add_particle_group(grow=0.1, max_age=2)
particles.add_color_stop(0, GREEN)
particles.add_color_stop(2, TRANSPARENT_GREEN)

player1 = make_player(
    pos=center - Vector2(200, 0),
    angle=-math.pi * 0.5
)
player1.score_label = score2
player2 = make_player(
    pos=center + Vector2(200, 0),
    angle=math.pi * 0.5
)
player2.score_label = score1

star = scene.layers[-1].add_circle(
    radius=40,
    fill=False,
    color=GREEN,
    stroke_width=LINE_W,
    pos=(scene.width / 2, scene.height / 2)
)

objects = [star, player1, player2]

joystick.init()
sticks = [joystick.Joystick(i) for i in range(min(joystick.get_count(), 2))]
for s in sticks:
    s.init()


def pressed(key):
    return lambda: keyboard[key]


controls = [
    (player1, pressed(keys.W), pressed(keys.A), pressed(keys.D), keys.E),
    (player2, pressed(keys.UP), pressed(keys.LEFT), pressed(keys.RIGHT), keys.RETURN),
]

def make_stick_controls(s):
    return (
        lambda: s.get_axis(1) < -0.5,
        lambda: s.get_axis(0) < -0.5,
        lambda: s.get_axis(0) > 0.5,
    )

for i, s in enumerate(sticks):
    player, *_, shoot = controls[i]
    controls[i] = (
        player,
        *make_stick_controls(s),
        shoot,
    )


def collides(a, b) -> bool:
    """Test if two objects have collided."""
    sep = a.pos - b.pos
    radii = a.radius + b.radius

    return Vector2(*sep).length_squared() < radii * radii


async def respawn(obj):
    obj.score_label.value += 1
    obj.score_label.text = str(obj.score_label.value)

    obj.dead = True
    obj.color = INVISIBLE
    await w2d.clock.coro.sleep(3)

    ring = scene.layers[0].add_circle(
        radius=40,
        pos=obj.initial_pos,
        color=TRANSPARENT_GREEN,
        fill=False,
        stroke_width=LINE_W,
    )
    w2d.tone.play(256, 1.0)
    await w2d.animate(
        ring,
        'accelerate',
        duration=0.5,
        scale=0.1,
        color=GREEN,
        stroke_width=10,
    )
    ring.delete()

    objects.append(obj)
    obj.dead = False
    obj.angle = obj.initial_angle
    obj.pos = obj.initial_pos
    obj.v = Vector2(obj.initial_v)
    for i in range(11):
        obj.color = INVISIBLE if i % 2 else GREEN
        await w2d.clock.coro.sleep(0.1)
        if obj.dead:
            return


def collision_pairs(objects):
    objects.sort(key=lambda o: o.pos[0] - o.radius)

    open = []
    for o in objects:
        x = o.pos[0]
        r = o.radius
        left = x - r
        new_open = []
        for o2r, o2 in open:
            if o2r < left:
                continue
            new_open.append((o2r, o2))
            if collides(o, o2):
                yield o, o2
        right = x + r
        open.append((right, o))


def update(dt):
    fps.text = f'FPS: {scene.fps:0.1f}'
    dt = min(dt, 0.5)
    dead = set()

    for o in objects:
        if o is star:
            continue

        sep = Vector2(*star.pos - o.pos)
        dist = sep.magnitude()
        if dist > 1500:
            # If it's flying off into space, kill it
            dead.add(o)
            o.silent = True
        o_u = o.v
        o.v += GRAVITY / (dist * dist) * sep.normalize() * dt
        o.pos += (o_u + o.v) * 0.5 * dt

    for a, b in collision_pairs(objects):
        dead |= {a, b}
    dead.discard(star)
    objects[:] = [o for o in objects if o not in dead]

    for o in dead:
        particles.emit(
            o.radius ** 2,
            size=1,
            pos=o.pos,
            pos_spread=o.radius * 0.7,
            vel=o.v * 0.2,
            vel_spread=50,
            color=GREEN
        )

    for obj, up, left, right, _ in controls:
        if obj in dead:
            dead.discard(obj)
            w2d.tone.play(20, 1.0, waveform='square')
            w2d.clock.coro.run(respawn(obj))
            continue
        elif obj.dead:
            continue

        if up():
            obj.v += forward(obj, ACCEL * dt)
            particles.emit(
                np.random.poisson(30 * dt),
                size=2,
                pos=obj.pos + forward(obj, -7),
                vel=obj.v + forward(obj, -100),
                vel_spread=4,
                color=GREEN
            )

        if left():
            obj.angle -= ROTATION_SPEED * dt
        elif right():
            obj.angle += ROTATION_SPEED * dt

    for o in dead:
        if not getattr(o, 'silent', False):
            w2d.tone.play(30, 0.3, waveform='square')
        o.delete()


@w2d.event
def on_key_down(key):
    if key == keys.ESCAPE:
        score1.text = score2.text = '0'
        score1.value = score2.value = 0
    elif key == keys.P:
        w2d.clock.default_clock.paused = not w2d.clock.default_clock.paused
        return

    for ship, _, _, _, shoot_button in controls:
        if shoot_button is key:
            break
    else:
        return
    shoot(ship)


def shoot(ship):
    if ship.dead:
        return

    bullet = scene.layers[0].add_rect(
        4, 4,
        pos=ship.pos + forward(ship, 12),
        color=GREEN
    )
    bullet.radius = 2.8

    v = forward(ship, 200)
    bullet.v = ship.v + v
    # Recoil!
    #ship.v -= v * 0.02

    objects.append(bullet)
    waveform = 'triangle' if ship is player1 else 'saw'
    w2d.tone.play(200, 0.3, waveform=waveform, volume=0.6)


@w2d.event
def on_joybutton_down(joy, button):
    if joy >= 2:
        return
    ship = controls[joy][0]
    shoot(ship)

w2d.clock.each_tick(update, strong=True)
w2d.run()



"""A group of other Transformables.

By grouping transformables they can be moved, rotated and scaled as one.

"""
from typing import List, Tuple

import numpy as np

from .base import Transformable


Point = Tuple[float, float]
Vector = Tuple[float, float]


class Group(Transformable):
    def __init__(
            self,
            transformables: List[Transformable],
            *,
            pos: Point = (0.0, 0.0),
            angle: float = 0.0,
            scale: float = 1.0):
        """Construct a group containing the given transformables."""
        super().__init__()
        self._objects = []
        self._result_mat = self._Transformable__xfmat.copy()
        self._inv_mat = None
        self.pos = pos
        self.angle = angle
        self.scale = scale
        self.extend(transformables)

    def extend(self, transformables: List[Transformable]):
        """Add transformables to the group."""
        for o in transformables:
            self._prepare(o)
            self._objects.append(o)

    def append(self, transformable: Transformable):
        """Add one transformable to the group."""
        self.extend([transformable])

    def _prepare(self, o):
        """Set up an object for membership of the group."""
        if o._group_xform is not None:
            raise ValueError(f"{o} is already in a group")
        elif not isinstance(o, Transformable):
            raise TypeError("Only Transformable objects can be grouped")
        o._group_xform = self._result_mat
        o._set_dirty()

    def _set_dirty(self):
        self._result_mat[:] = self._xform()
        self._inv_mat = None
        for o in self._objects:
            o._set_dirty()

    def __getitem__(self, idx: int) -> Transformable:
        """Get the transformable at the given position."""
        return self._objects[idx]

    def __delitem__(self, idx: int) -> Transformable:
        """Delete the transformable at the given position."""
        obj = self._objects.pop(idx)
        obj.delete()

    def __setitem__(self, idx: int, o: Transformable) -> Transformable:
        """Replace the object at the given index."""
        self._prepare(o)
        self._objects[idx].delete()
        self._objects[idx] = o

    def __len__(self):
        """Get the number of objects in the group."""
        return len(self._objects)

    def local_to_world(self, point: Point) -> Point:
        """Get the world position for the given local point."""
        x, y = point
        return np.array([x, y, 1]) @ self._result_mat[:, :2]

    def world_to_local(self, point: Point) -> Point:
        """Get the world position for the given local point."""
        x, y = point
        return np.array([x, y, 1]) @ self._inv[:, :2]

    def localvec_to_worldvec(self, vec: Vector) -> Vector:
        """Get the world vector for the given local vector."""
        x, y = vec
        return np.array([x, y, 0]) @ self._result_mat[:, :2]

    def worldvec_to_localvec(self, vec: Vector) -> Vector:
        """Get the world vector for the given local vector."""
        x, y = vec
        return np.array([x, y, 0]) @ self._inv[:, :2]

    @property
    def _inv(self):
        """Get the inverse transformation matrix, memoized."""
        if self._inv_mat is None:
            self._inv_mat = np.linalg.inv(self._result_mat)
        return self._inv_mat

    def delete(self):
        """Remove all primitives in the group."""
        for o in self._objects:
            o.delete()
        del self.objects[:]

    #: Alias for delete to match interface of a list
    clear = delete

    def pop(self, index: int) -> Transformable:
        """Extract a primitive from the group, leaving it in the scene.

        The primitive will have (approximately) the same world transformation
        it had within the group.
        """
        obj = self._objects.pop(index)
        mat = obj._xform()
        obj._group_xform = None
        self._factorise(obj, mat)

        # We cannot represent the skew from a non-linear scale, so this
        # operation actually does modify the matrix. If it didn't then the
        # object wouldn't be dirty.
        obj._set_dirty()
        return obj

    def explode(self) -> List[Transformable]:
        """Extract and return primitives from the group.

        This preserves the transformation for the primitives and leaves them
        in the scene.

        This group becomes empty after this operation.
        """
        # We can repeatedly call pop(), but we do it in reverse order because
        # otherwise it's quadratic
        num = len(self._objects)
        items = [None] * num
        for i in range(num - 1, -1, -1):
            items[i] = self.pop(i)
        return items

    def _factorise(self, obj, mat):
        """Assign an object's transformation properties from mat."""
        x = mat[:2, 0]
        y = mat[:2, 1]
        obj._Transformable__xfmat = mat  # sets position
        obj.angle = np.arctan2(y[0], x[0])
        obj.scale_x = np.sqrt(np.sum(x * x))
        obj.scale_y = np.sqrt(np.sum(y * y))

    def capture(self, *objects: Tuple[Transformable]):
        """Capture the given objects into the group.

        The objects will have the same transformation they had outside of the
        group.

        """
        inv = self._inv
        for o in objects:
            mat = o._xform() @ inv
            self._prepare(o)
            self._objects.append(o)
            self._factorise(o, mat)

    @classmethod
    def from_objects(cls, objects: List[Transformable], **kwargs) -> 'Group':
        """Build a group by capturing the given objects.

        Keyword arguments can be used to set the initial transform for the
        group prior to capturing the objects.
        """
        group = cls([], **kwargs)
        group.capture(*objects)
        return group



from typing import Tuple, Any, Optional
from dataclasses import dataclass

import numpy as np
import numpy.random
from sortedcontainers import SortedList

from .base import Transformable
from ..clock import default_clock
from ..color import convert_color
from ..allocators.vertlists import VAO


PARTICLE_DTYPE = np.dtype([
    ('in_vert', '2f4'),
    ('in_color', '4f4'),
    ('in_age', 'f4'),
    ('in_size', 'f4'),
    ('in_angle', 'f4'),
])


class ParticleVAO(VAO):
    """A VAO with an image texture and a color ramp texture."""

    def __init__(self, pgroup, *args, **kwargs):
        super().__init__(*args, dtype=PARTICLE_DTYPE, **kwargs)
        self.pgroup = pgroup

    def render(self, camera):
        self.prog['grow'].value = self.pgroup.grow
        self.prog['max_age'].value = self.pgroup.max_age
        self.prog['tex'].value = 0
        self.prog['color_tex'].value = 1
        self.tex.use(0)
        self.color_tex.use(1)
        super().render(camera)


class ParticleGroup:
    """A group of particles."""

    def __init__(
            self,
            layer,
            clock=default_clock,
            *,
            grow: float = 1.0,
            max_age: float = np.inf,
            gravity: Tuple[float, float] = (0, 0),
            drag: float = 1.0,
            spin_drag: float = 1.0):
        super().__init__()
        self.layer = layer
        self.num: int = 0  # Number of particles we have
        self.max_age = max_age
        self.grow = grow
        self.gravity = np.array(gravity)
        self.drag = drag
        self.spin_drag = spin_drag
        self.spins = np.zeros([0])
        self.vels = np.zeros([0, 2])
        self._color_stops = SortedList()
        self.color_tex = layer.ctx.texture((512, 1), 4, dtype='f2')
        self.color_vals = np.ones((512, 4), dtype='f2')
        self.color_tex.write(self.color_vals)
        self.emitters = set()
        self._clock = clock
        clock.each_tick(self._update)

    def add_color_stop(self, age, color):
        """Add a color stop for particles of the given age.

        Particles will fade between the colors of the stops as their age
        changes.
        """
        color = convert_color(color)
        self._color_stops.add((age, *color))
        xs = np.linspace(0, self.max_age, 512)

        ages, *colors = zip(*self._color_stops)
        for i in range(4):
            self.color_vals[:, i] = np.interp(
                xs,
                ages,
                colors[i],
            )
        self.color_tex.write(self.color_vals)

    def emit(
            self,
            num: int,
            *,
            pos: Tuple[float, float],
            pos_spread: float = 0,
            vel: Tuple[float, float] = (0, 0),
            vel_spread: float = 0,
            color: Any = (1, 1, 1, 1),
            size: float = 1.0,
            size_spread: float = 0.0,
            spin: float = 0.0,
            spin_spread: float = 0.0,
            angle: float = 0.0,
            angle_spread: float = 0.0):
        """Emit num particles."""
        num = round(num)
        if num == 0:
            return
        color = convert_color(color)

        prev_verts = self.lst.vertbuf
        alive = self.lst.vertbuf['in_age'] < self.max_age
        num_alive = np.sum(alive)
        need = num_alive + num

        verts_alive = prev_verts[alive]

        self.lst.realloc(need, need)
        self.lst.indexbuf[:] = np.arange(need, dtype='u4') + self.lst.vertoff.start

        new_vel = np.random.normal(vel, vel_spread, [num, 2])
        new_pos = np.random.normal(pos, pos_spread, [num, 2])
        new_size = np.random.normal(size, size_spread, num)
        new_angles = np.random.normal(angle, angle_spread, num)
        new_spins = np.random.normal(spin, spin_spread, num)

        verts = self.lst.vertbuf
        self.vels = np.vstack([self.vels[alive], new_vel])
        self.spins = np.hstack([self.spins[alive], new_spins])
        verts[:num_alive] = verts_alive
        new = verts[num_alive:]
        new['in_age'] = 0
        new['in_color'] = color
        new['in_size'] = new_size
        new['in_vert'] = new_pos
        new['in_angle'] = new_angles
        self.lst.dirty = True

    def _compact(self):
        alive = self.lst.vertbuf['in_age'] < self.max_age
        self.num = num_alive = np.sum(alive)
        verts_alive = self.lst.vertbuf[alive]
        self.lst.realloc(num_alive, num_alive)

        first_vertex = self.lst.vertoff.start
        self.lst.indexbuf[:] = np.arange(
            first_vertex,
            first_vertex + num_alive,
            dtype='u4'
        )
        self.vels = self.vels[alive].copy()
        self.spins = self.spins[alive].copy()
        self.lst.vertbuf[:] = verts_alive

    def _update(self, dt):
        self.lst.vertbuf['in_age'] += dt
        self._compact()

        # Update
        orig_vels = self.vels
        self.vels *= self.drag ** dt
        self.vels += self.gravity * dt
        self.spins *= self.spin_drag ** dt

        self.lst.vertbuf['in_vert'] += (self.vels + orig_vels) * (dt * 0.5)
        self.lst.vertbuf['in_angle'] += self.spins * dt
        self.lst.dirty = True

        for e in self.emitters:
            e._emit(dt)

    def _migrate(self, vao: VAO):
        """Migrate the particles into the given VAO."""
        self.vao = vao
        num = max(self.num, 1024)

        # Allocate a large slice (set high water mark)
        self.lst = vao.alloc(num, num)
        first_vertex = self.lst.vertoff.start
        self.lst.indexbuf[:] = np.arange(
            first_vertex,
            first_vertex + num,
            dtype='u4'
        )

        # Realloc to how much we actually want
        self.lst.realloc(self.num, self.num)

    def add_emitter(self, **kwargs):
        """Add a particle emitter object."""
        e = Emitter(self, **kwargs)
        self.emitters.add(e)
        return e

    def delete(self):
        self.layer.objects.discard(self)
        self._clock.unschedule(self._update)
        self.lst.free()


@dataclass
class EmitterDesc:
    """Descriptor for an emitter property."""
    default: Any = 0.0
    name: Optional[str] = None

    def __set_name__(self, cls, name):
        if self.name is None:
            self.name = name

    def __get__(self, inst, cls):
        return inst._params.get(self.name, self.default)

    def __set__(self, inst, value):
        inst._params[self.name] = value


class Emitter(Transformable):
    """A transformable emitter object."""

    group: ParticleGroup
    pos_spread: float = EmitterDesc(0)
    vel: Tuple[float, float] = (0, 0)
    vel_spread: float = EmitterDesc(0)
    color: Any = EmitterDesc((1, 1, 1, 1))
    size: float = EmitterDesc(1.0)
    size_spread: float = EmitterDesc(0.0)
    spin: float = EmitterDesc(0.0)
    spin_spread: float = EmitterDesc(0.0)
    emit_angle: float = 0.0
    emit_angle_spread: float = EmitterDesc(0.0, 'angle_spread')

    def __init__(
        self,
        group: ParticleGroup,
        *,
        rate: float = 100.0,
        pos: Tuple[float, float] = (0, 0),
        angle: float = 0.0,
        scale: float = 1.0,
        **kwargs,
    ):
        super().__init__()
        self._group = group
        self._params = {}
        self.rate = rate

        self.pos = pos
        self.angle = angle
        self.scale = scale

        self._vecs = np.array([
            [0, 0, 1],
            [*self.vel, 0],
            [1, 0, 0],
        ], dtype=np.float32)

        for k, v in kwargs.items():
            if k not in type(self).__dict__:
                raise TypeError(
                    f"{type(self).__name__}.__init__() does not accept a "
                    f"keyword argument {k}!r"
                )
            setattr(self, k, v)

    def delete(self):
        """Delete the emitter."""
        self._group.emitters.discard(self)

    def _set_dirty(self):
        """We don't need to track dirtyiness."""

    def _emit(self, dt):
        """Emit particles in the group."""
        num = np.random.poisson(self.rate * dt)

        if num == 0:
            return

        xform = self._xform()[:, :2]

        self._vecs[1, :2] = self.vel
        pos, vel, (rotx, roty) = self._vecs @ xform
        angle = np.arctan2(roty, rotx) + self.emit_angle

        self._group.emit(
            num,
            pos=pos,
            vel=vel,
            angle=angle,
            **self._params
        )



import unicodedata
import pkgutil
import io

import moderngl
import numpy as np
import pygame.font

from ..allocators.vertlists import VAO
from ..loaders import fonts
from ..atlas import Atlas
from .sprites import QUAD
from .base import Bounds, Transformable, Colorable


FONT_LOAD_SIZE = 48
default_font = None


def get_default_font():
    """Get the default font."""
    global default_font
    if not default_font:
        ttf_data = pkgutil.get_data('wasabi2d', 'data/roboto_regular.ttf')
        default_font = pygame.font.Font(io.BytesIO(ttf_data), FONT_LOAD_SIZE)
    return default_font


class TextureVAO(VAO):
    """A VAO with an associated texture."""

    def render(self, camera):
        self.prog['tex'].value = 0
        self.tex.use(0)
        super().render(camera)


def text_vao(
        ctx: moderngl.Context,
        shadermgr: 'wasabi2d.layers.ShaderManager') -> VAO:
    """Build a BAO for rendering plain colored objects."""
    return TextureVAO(
        mode=moderngl.TRIANGLES,
        ctx=ctx,
        prog=shadermgr.load('texquads', 'text'),
        dtype=np.dtype([
            ('in_vert', '2f4'),
            ('in_color', '4f4'),
            ('in_uv', '2u2'),
        ])
    )


class FontAtlas(Atlas):
    """The combination of a font and the texture atlas it uses."""
    def __init__(self, ctx, font_name):
        super().__init__(ctx)
        if font_name is None:
            self.font = get_default_font()
        else:
            self.font = fonts.load(font_name, fontsize=FONT_LOAD_SIZE)

    def _load(self, name):
        """Load the image for the given name."""
        return self.font.render(name, True, (255, 255, 255))


# Text alignments as a fraction of text layout width
ALIGNMENTS = {
    'left': 0,
    'center': 0.5,
    'right': 1,
}


class Label(Colorable, Transformable):
    """A single-line text block with no additional layout/wrapping."""

    def __init__(
            self,
            text: str,
            font_atlas: FontAtlas,
            layer,
            *,
            align='left',
            fontsize=20,
            pos=(0, 0),
            color=(1, 1, 1, 1)):
        assert align in ALIGNMENTS, f"{align!r} is not a valid text alignment."
        super().__init__()
        self.layer = layer
        self.lst = None
        self._verts = None
        self.align = align
        self.layer = layer
        self.fontsize = fontsize
        self.font_atlas = font_atlas
        self.pos = pos
        self.color = color
        self.vao = None
        self.text = text  # trigger layout

    def _set_dirty(self):
        self.layer._dirty.add(self)

    @property
    def align(self):
        """Get the alignment of the text relative to pos."""
        return self._align

    @align.setter
    def align(self, align):
        if align not in ALIGNMENTS:
            raise ValueError(f"{align!r} is not a valid text alignment.")
        self._align = align
        if self._verts is not None:
            self._layout()

    @property
    def text(self) -> str:
        """Get the text of this label."""
        return self._text

    @text.setter
    def text(self, text: str):
        """Set the text of this label.

        The string is currently normalised to composed form because this
        is simpler to lay out than a string with combining characters, and
        we may not be very sophisticated at this point.

        """
        self._str = unicodedata.normalize('NFC', str(text))
        self._text = text
        self._layout()

        self._set_dirty()

    def _layout(self):
        """Generate indices, uvs and verts for the text."""
        font = self.font_atlas.font

        descent = font.get_descent()
        n_chars = len(self._str) - self._str.count('\n')
        verts = np.ones((4 * n_chars, 3), dtype='f4')
        uvs = np.zeros((4 * n_chars, 2), dtype='f4')
        indices = np.zeros(n_chars * 6, dtype='u4')

        lines = self._str.split('\n')

        # We lay out based on 48px tex so line size is scaled later
        line_height = 48 * 1.3

        curchar = 0
        tex_ids = set()
        tex = None

        for lineno, line in enumerate(lines):
            if not line:
                continue
            # (min_x, max_x, min_y, max_y, horizontal_advance_x)
            metrics = np.array(font.metrics(line), dtype='f4')
            cx = np.cumsum(metrics[:, 4]).reshape(-1, 1)
            xpos = metrics[:, 0:2] + cx

            layout_width = cx[-1] + metrics[-1, 1]
            align_offset = ALIGNMENTS[self._align] * layout_width
            yoff = lineno * line_height
            for idx, char in enumerate(line):
                texregion = self.font_atlas.get(char)
                # TODO: this could break if tex is reallocated (eg. because it
                # grows)
                tex = texregion.tex.tex
                glyph_uvs = texregion.texcoords
                glyph_verts = texregion.get_verts(texregion.width, texregion.height)
                tex_ids.add(tex.glo)

                # The kerning seems pretty bad on Pygame fonts...
    #            glyph_width = glyph_verts[1, 0] - glyph_verts[0, 0]
    #            metrics_width = xpos[idx, 1] - xpos[idx, 0]
    #            print(repr(char), glyph_width, metrics_width, metrics[idx, 4])

                x = xpos[idx, 0]

                quadnum = idx + curchar
                start = quadnum * 4
                glyph_slice = slice(start, start + 4)
                verts[glyph_slice] = glyph_verts + (x - align_offset, yoff - descent, 0)
                uvs[glyph_slice] = glyph_uvs
                indices[6 * quadnum:6 * quadnum + 6] = QUAD + 4 * quadnum

            curchar += len(line)

        # Scale coordinates
        resize = np.identity(3, dtype='f4')
        scale = self.fontsize / 48
        resize[0, 0] = scale
        resize[1, 1] = scale

        # TODO: handle use of multiple textures. We will be able to handle
        # this eventually by selecting texture unit within the shader, or by
        # making multiple draw calls
        assert len(tex_ids) < 2, "Label got allocated over multiple textures"
        self.tex = tex
        self._verts = verts @ resize
        self._uvs = uvs
        self._indices = indices

        # TODO: update self.lst, set dirty OR reallocate self.lst for new size
        if self.tex is None:
            if self.lst:
                self.lst.free()
                self.lst = None
                self.vao = None
        elif self.lst:
            self.lst.realloc(len(self._verts), len(indices))
            self.lst.indexbuf[:] = indices
            self.lst.indexbuf += self.lst.vertoff.start
            self.lst.vertbuf['in_uv'] = uvs
            self._update()
        elif self.tex and not self.vao:
            self._migrate(self.layer._text_vao(self.font_atlas))

    bounds = Bounds('self._verts[:, :2]')

    def delete(self):
        self.layer.objects.remove(self)
        self.lst.free()
        self.layer = None
        self.vao = None

    def _update(self):
        if not self.lst:
            return
        xform = self._xform()

        np.matmul(
            self._verts,
            xform[:, :2],
            self.lst.vertbuf['in_vert']
        )
        self.lst.vertbuf['in_color'] = self._color
        self.lst.dirty = True

    def _migrate(self, vao: TextureVAO):
        """Migrate the fill into the given VAO."""
        # TODO: dealloc from an existing VAO
        if self.lst:
            self.lst.free()
            self.lst = None
        idxs = self._indices
        self.vao = vao
        self.vao.tex = self.tex
        self.lst = vao.alloc(len(self._verts), len(idxs))
        self.lst.indexbuf[:] = idxs
        self.lst.indexbuf += self.lst.vertoff.start
        self.lst.vertbuf['in_uv'] = self._uvs
        self._update()



"""Example of the sepia effect."""
import wasabi2d as w2d

scene = w2d.Scene()

photo = scene.layers[0].add_sprite(
    'positano',
    pos=(scene.width / 2, scene.height / 2),
)
photo.scale = max(
    scene.width / photo.width,
    scene.height / photo.height
)
sepia = scene.layers[0].set_effect('sepia')


@w2d.event
def on_mouse_move(pos):
    x, y = pos
    frac = x / scene.width
    sepia.amount = frac


w2d.run()



"""Example of using one layer to lens another."""
import wasabi2d as w2d
import pygame.surface
import pygame.surfarray
import pygame.image
from pathlib import Path
import numpy as np

scene = w2d.Scene()
center = (scene.width / 2, scene.height / 2)

photo = scene.layers[0].add_sprite(
    'positano',
    pos=center
)
photo.scale = max(
    scene.width / photo.width,
    scene.height / photo.height
)


def make_lens():
    sz = 800
    surf = pygame.Surface((sz, sz), depth=32, flags=pygame.SRCALPHA)
    center = sz * 0.5
    for y in range(sz):
        for x in range(sz):
            off = np.array([x, sz - y]) - center
            off /= center

            dist = np.sum(off * off)
            color = (
                *np.clip(127.5 - 64 * off - 64 * off * dist, 0, 255),
                dist,
                0 if dist > 1 else 255
            )
            surf.set_at((x, y), color)
    surf = pygame.transform.smoothscale(surf, (200, 200))
    pygame.image.save(surf, str(Path(__file__).parent / 'images/lens.png'))


#make_lens()

lens = scene.layers[1].add_sprite(
    'lens_8x',
    pos=center,
)

scene.chain = [
    w2d.chain.LayerRange(stop=0),
    w2d.chain.DisplacementMap(
        displacement=w2d.chain.Layers([1]),
        paint=w2d.chain.LayerRange(stop=0),
        scale=400
    )
]


@w2d.event
def on_mouse_move(pos):
    lens.pos = pos


w2d.run()



"""Example of using one layer to mask another."""
import wasabi2d as w2d
import math
from pygame import Rect
from wasabi2d import Vector2
from wasabi2d.actor import Actor

TILE = 21
TILES_W = 15
TILES_H = 10

grid = set()
grid.update((x, TILES_H) for x in range(TILES_W))
grid.update((-1, y) for y in range(-TILES_H, TILES_H))
grid.update((TILES_W, y) for y in range(-TILES_H, TILES_H))

scene = w2d.Scene(
    width=TILE * TILES_W,
    height=TILE * TILES_H,
    scaler=True,
    pixel_art=True,
)
scene.background = '#5e81a2'
scene.layers[1].set_effect('dropshadow', radius=2, offset=(0, 1))

alien = scene.layers[1].add_sprite(
    'pc_standing',
    anchor_x=10,
    anchor_y=21,
    pos=(210, TILE * 9)
)
alien.fpos = Vector2(*alien.pos)
alien.v = Vector2(0, 0)
alien.stood = True
alien.crouch = False


def create_platform(x1, x2, y):
    length = x2 - x1
    if length == 1:
        grid.add((x1, y))
        scene.layers[1].add_sprite(
            'platform_single',
            pos=(x1 * TILE, y * TILE),
            anchor_x=0,
            anchor_y=0,
        )
        return
    for i in range(length):
        pos = x1 + i, y
        grid.add(pos)
        if i == 0:
            sprite = 'platform_l'
        elif i == (length - 1):
            sprite = 'platform_r'
        else:
            sprite = 'platform_m'

        scene.layers[1].add_sprite(
            sprite,
            pos=((x1 + i) * TILE, (y * TILE)),
            anchor_x=0,
            anchor_y=0,
        )


create_platform(0, 15, 9)
create_platform(12, 14, 8)
create_platform(3, 6, 6)

ACCEL = 0.7
JUMP = 7.7
GRAVITY = 0.5
DRAG = 0.8


def world_to_grid(pos):
    x, y = pos
    return x / TILE, y / TILE


def collide_point(*pos):
    """Is the given world coordinate in the grid."""
    pos = {(x // TILE, y // TILE) for x, y in pos}
    return bool(pos & grid)


def tile_floor(val):
    return math.floor(val / TILE) * TILE


def tile_ceil(val):
    return math.ceil(val / TILE) * TILE


eps = 1e-4
BOUNDS = [
    Vector2(11 - eps, -eps),  # br
    Vector2(-10 + eps, -eps),  # bl
    Vector2(-10 + eps, -21 + eps),  # tl
    Vector2(11 - eps, -21 + eps),  # tr
]


def bounds():
    """Get the alien's bounds as a tuple (br, bl, tl, tr)."""
    pos = alien.fpos
    return tuple(pos + v for v in BOUNDS)


@w2d.event
def update(keyboard):
    x, y = alien.fpos
    vx, vy = alien.v

    y += vy
    alien.fpos.y = y

    eps = 1e-4

    br, bl, tl, tr = bounds()

    if alien.stood:
        belowl = bl + Vector2(0, 1)
        belowr = br + Vector2(0, 1)
        if not (collide_point(belowl, belowr)):
            alien.stood = False
            alien.image = 'pc_standing'
            vy += GRAVITY
    else:
        if vy + eps > y - tile_floor(y) > 0:
            if collide_point(bl, br):
                alien.image = 'pc_standing'
                alien.stood = True
                vy = 0
                y = tile_floor(y)
            else:
                alien.image = 'pc_falling'
        elif vy - eps < tl.y - tile_ceil(tl.y) < 0:
            if collide_point(tl, tr):
                print("oof")
                vy = 0
                y = tile_ceil(y)
        if not alien.stood:
            vy += GRAVITY

    if not alien.crouch:
        if keyboard.left:
            vx -= ACCEL
        elif keyboard.right:
            vx += ACCEL
    vx *= DRAG
    x += vx
    alien.fpos.x = x

    br, bl, tl, tr = bounds()

    if vx + eps > tr.x - tile_floor(tr.x) > 0:
        alien.scale_x = 1
        if collide_point(tr, br):
            vx = 0
            x = tile_floor(tr.x) - 11
    elif vx - eps < tl.x - tile_ceil(tl.x) < 0:
        alien.scale_x = -1
        if collide_point(tl, bl):
            vx = 0
            x = tile_ceil(tl.x) + 10
    alien.fpos.x = vx

    alien.v = Vector2(vx, vy)
    alien.pos = alien.fpos = Vector2(x, y)



@w2d.event
def on_key_down(key):
    if key is w2d.keys.UP:
        if alien.stood and not alien.crouch:
            alien.v.y = -JUMP
            alien.stood = False
            alien.image = 'pc_jumping'
    elif key is w2d.keys.DOWN:
        if alien.stood:
            print("crouch")
            alien.image = 'pc_crouch'
            alien.crouch = True


@w2d.event
def on_key_up(key):
    if key is w2d.keys.DOWN:
        if alien.crouch:
            alien.image = 'pc_standing'
            alien.crouch = False

w2d.run()



"""Pack sprites into texture atlases."""
from typing import Tuple, Optional, List, Union
from dataclasses import dataclass

import moderngl
import numpy as np
import pygame
from pygame import Rect

from .loaders import images
from .allocators.textures import Packer, NoFit
from .primitives.sprites import QUAD
from .shaders import shadermgr, blend_func, bind_framebuffer


BLIT_PROGRAM = dict(
    vertex_shader='''
        #version 330

        in ivec2 in_uv;
        uniform ivec2 newsize;
        out vec2 uv;
        uniform sampler2D tex;

        const vec2 BL = vec2(-1, -1);

        void main() {
            vec2 v = vec2(in_uv) / newsize;
            gl_Position = vec4(v * 2 + BL, 0.0, 1.0);
            uv = vec2(in_uv) / textureSize(tex, 0);
        }
    ''',
    fragment_shader='''
        #version 330

        out vec4 f_color;
        in vec2 uv;
        uniform sampler2D tex;

        void main() {
            f_color = texture(tex, uv);
        }
    ''',
)


class TexSurface:
    """A GPU texture.

    This class allows copying pygame Surfaces into regions of a texture on the
    GPU.
    """
    ctx: moderngl.Context
    tex: moderngl.Texture
    _dirty: bool

    def __init__(self, ctx: moderngl.Context, tex: moderngl.Texture):
        self.ctx = ctx
        self.tex = tex
        self._dirty = False

    @property
    def width(self):
        """Get the current width of the texture."""
        return self.tex.width

    @property
    def height(self):
        """Get the current height of the texture."""
        return self.tex.height

    @classmethod
    def new(cls, ctx: moderngl.Context, size: Tuple[int, int]) -> 'TexSurface':
        """Create a new TexSurface in the given moderngl context."""
        tex = ctx.texture(size, 4)
        return cls(ctx, tex)

    def resize(self, newsize: Tuple[int, int]):
        """Resize the texture and copy the existing data into it."""
        newtex = self.ctx.texture(newsize, 4)
        in_region = TextureRegion.for_tex(self.tex)
        vdata = in_region.texcoords

        vs = self.ctx.buffer(vdata)
        ibuf = self.ctx.buffer(QUAD)
        prog = shadermgr(self.ctx).get(**BLIT_PROGRAM)
        self.tex.use(0)
        self.tex.filter = moderngl.NEAREST, moderngl.NEAREST
        prog['tex'].value = 0
        prog['newsize'].value = newsize
        vao = self.ctx.vertex_array(
            prog,
            [
                (vs, '2u2', 'in_uv'),
            ],
            ibuf
        )
        fb = self.ctx.framebuffer(color_attachments=[newtex])
        with bind_framebuffer(self.ctx, fb), \
                blend_func(self.ctx, moderngl.ONE, moderngl.ZERO):
            vao.render(vertices=6)
        self._dirty = True
        self.tex.release()
        self.tex = newtex

    def write(self, img: pygame.Surface, rect: pygame.Rect):
        """Write the contents of img at the given coordinates."""
        imgdata = pygame.image.tostring(img, "RGBA", 1)
        self.tex.write(imgdata, (rect.x, rect.y, rect.width, rect.height))
        self._dirty = True

    def update(self):
        """Sync texture data to the GPU."""
        self.tex.build_mipmaps(max_level=2)
        self.tex.filter = self.ctx.extra['texture_filter']
        self._dirty = False

    def save(self, fname: str):
        """Save the contents of the texture to a file.

        This is intended for debugging.
        """
        w, h = self.tex.width, self.tex.height
        data = self.tex.read(components=4)
        assert len(data) == (w * h * 5), \
            f"Received {len(data)}, expected {w * h * 4}"
        img = pygame.image.fromstring(data, (w, h), 'RGBA')
        img = pygame.transform.flip(img, False, True)
        pygame.image.save(img, fname)

    def use(self, texunit: int):
        """Shortcut to bind the texture to a texture unit."""
        self.tex.use(texunit)

    def __del__(self):
        self.tex.release()


@dataclass
class TextureRegion:
    """An allocated region of a texture."""
    tex: TexSurface
    width: int
    height: int
    texcoords: np.ndarray
    rot: int = 0

    def __post_init__(self):
        if not isinstance(self.rot, int):
            raise TypeError(f"rot must be int, not {self.rot!r}")
        if self.texcoords.dtype != np.uint16:
            raise TypeError(
                f"Invalid dtype {self.texcoords.dtype} for tex coords"
            )
        w = self.width
        h = self.height
        self.verts = np.array([
            (0, 0, 1),
            (w, 0, 1),
            (w, h, 1),
            (0, h, 1),
        ], dtype='f4')

    @classmethod
    def for_tex(cls, tex):
        """Create a TextureRegion corresponding to the whole of tex."""
        return cls.for_rect(tex, Rect(0, 0, tex.width, tex.height))

    def absregion(self) -> Rect:
        """Get the region of the original texture."""
        (l, t), (r, b) = self.texcoords[[0, 2]].astype(int)
        w = abs(l - r)
        h = abs(b - t)
        l = min(l, r)
        t = min(b, t)
        return Rect(l, t, w, h)

    @classmethod
    def for_rect(cls, tex, rect: Rect):
        """Create a TextureRegion for the given texture and rect."""
        if isinstance(tex, TextureRegion):
            assert rect.width and rect.height, "Invalid rect dimensions"
            myrect = Rect(0, 0, tex.width, tex.height)
            assert myrect.contains(rect), "Subrect is not in bounds."

            coords = tex.texcoords.astype(np.int32)
            lb = coords[3]
            r = np.sign(coords[2] - lb)
            u = np.sign(coords[0] - lb)
            lb += r * rect.left + u * rect.top
            across = r * rect.width
            up = u * rect.height
            texcoords = np.array([
                lb + up,
                lb + up + across,
                lb + across,
                lb,
            ], dtype=np.uint16)
            rot = tex.rot
            tex = tex.tex
        else:
            rot = 0

            l = rect.left
            b = rect.top
            r = rect.right
            t = rect.bottom
            texcoords = np.array([
                (l, t),
                (r, t),
                (r, b),
                (l, b),
            ], dtype=np.uint16)
        return cls(
            tex,
            rect.width,
            rect.height,
            texcoords,
            rot
        )

    def write(self, img: pygame.Surface):
        """Write the given surface into this region."""
        if self.rot:
            img = pygame.transform.rotate(img, self.rot)
        sz = img.get_size()
        bounds = self.absregion()
        assert sz == bounds.size, f"{sz!r} != {bounds.size!r}"
        self.tex.write(img, bounds)

    def rotated(self):
        """Get the view of this texture region rotated by 90 degrees."""
        newcoords = self.texcoords[[1, 2, 3, 0]]
        return TextureRegion(
            self.tex,
            self.height,
            self.width,
            newcoords,
            self.rot - 90,
        )

    ANCHOR_X_NAMES = {
        'left': 0,
        'center': 0.5,
        'right': 1,
    }
    ANCHOR_Y_NAMES = {
        'top': 0,
        'center': 0.5,
        'bottom': 1,
    }

    def get_verts(
        self,
        anchor_x: Union[float, str] = 'center',
        anchor_y: Union[float, str] = 'center',
    ) -> np.ndarray:
        """Get an array or transformed vertices."""
        if isinstance(anchor_x, str):
            anchor_x = self.width * self.ANCHOR_X_NAMES[anchor_x]
        if isinstance(anchor_y, str):
            anchor_y = self.height * self.ANCHOR_Y_NAMES[anchor_y]
        offset = (float(anchor_x), float(anchor_y), 0.0)
        vs = self.verts.copy()
        vs -= offset
        return vs


class Atlas:
    """Texture atlas generator.

    Newly requested textures are positioned using a packer and then written
    to a texture using a TexSurface.

    """

    def __init__(self, ctx, *, texsize=512, padding=2):
        self.ctx = ctx
        self.padding = padding
        self.texsize = texsize

        self.packer = Packer.new_shelves(size=texsize)

        self.texsurf: Optional[TexSurface] = None
        self.surfs_texs: List[TextureRegion] = []
        self.tex_for_name = {}

    def _load(self, name):
        """Load the image for the given name."""
        return images.load(name)

    def mksurftex(self) -> TextureRegion:
        """Make a new surface and corresponding texture of the given size."""
        if not self.texsurf:
            self.texsurf = TexSurface.new(self.ctx, (self.texsize,) * 2)
            region = TextureRegion.for_tex(self.texsurf)
        else:
            count = len(self.surfs_texs)
            self.texsurf.resize((
                (count + 1) * self.texsize,
                self.texsize
            ))
            region = TextureRegion.for_rect(
                self.texsurf,
                Rect(count * self.texsize, 0, self.texsize, self.texsize)
            )

        self.surfs_texs.append(region)
        return region

    def npot_tex(self, sprite_name, img):
        """Get a non-power-of-two texture for this image."""
        w, h = size = img.get_size()
        tex = self.ctx.texture(size, 4)
        tex.write(pygame.image.tostring(img, "RGBA", 1))
        tex.build_mipmaps()
        tex.filter = self.ctx.extra['texture_filter']
        tex.repeat_x = tex.repeat_y = False
        texcoords = np.array([
            (0, h),
            (w, h),
            (w, 0),
            (0, 0),
        ], dtype=np.uint16)

        texregion = TextureRegion(TexSurface(self.ctx, tex), w, h, texcoords)
        res = self.tex_for_name[sprite_name] = texregion
        return res

    def get(self, sprite_name):
        if sprite_name in self.tex_for_name:
            return self.tex_for_name[sprite_name]

        img = self._load(sprite_name)

        pad = self.padding * 2

        orig = img.get_rect()
        r = Rect(orig)
        r.w += pad
        r.h += pad
        try:
            bin, packed = self.packer.add(r)
        except NoFit:
            # Does not fit in packer bounds
            return self.npot_tex(sprite_name, img)

        p = Rect(packed)
        p.left += self.padding
        p.top += self.padding
        p.w -= pad
        p.h -= pad

        try:
            reg = self.surfs_texs[bin]
        except IndexError:
            reg = self.mksurftex()

        texregion = TextureRegion.for_rect(reg, p)
        if orig.w != p.w:
            texregion = texregion.rotated()
        texregion.write(img)

        res = self.tex_for_name[sprite_name] = texregion
        return res

    def _update(self):
        """Copy updated surfaces to the GL texture objects."""
        if self.texsurf and self.texsurf._dirty:
            self.texsurf.update()

    def dump(self):
        tex = self.texsurf.tex
        data = tex.read()
        img = pygame.image.fromstring(data, (tex.width, tex.height), 'RGBA')
        img = pygame.transform.flip(img, False, True)
        pygame.image.save(img, 'atlas.png')



"""Pinball collision demo.

"""
import wasabi2d as w2d
import random
from pygame.math import Vector2 as v2
from pygame import Rect

from spatial_hash import SpatialHash


scene = w2d.Scene(1280, 720)
scene.chain = [
    w2d.chain.LayerRange(stop=0),
    w2d.chain.DisplacementMap(
        displacement=w2d.chain.Layers([1]),
        paint=w2d.chain.LayerRange(stop=0),
        scale=-100
    )
]

scene.layers[0].set_effect('dropshadow', offset=(3, 3))

cursor = scene.layers[0].add_sprite(
    'cursor',
    anchor_x='left',
    anchor_y='top'
)

scene.layers[-1].add_sprite('wood', anchor_x='left', anchor_y='top')

GRAVITY = v2(0, 1)
BALL_RADIUS = 15
BALL_COLOR = (34, 128, 75)
ELASTICITY = 0.3
SEPARATION_STEPS = [1.0] * 10

BALL_COUNT = 20


class Ball:
    def __init__(self, x, y, radius=15):
        self.sprite = scene.layers[0].add_sprite(
            'steel',
            scale=radius / 32,
        )
        self.refl = scene.layers[1].add_sprite(
            'lens_fresnel',
            scale=radius / 100,
            color=(1, 1, 1, 0.5)
        )
        self.velocity = v2(0, 0)
        self.radius = radius
        self.mass = radius * radius
        dr = v2(radius, radius)
        self.rect = Rect(0, 0, *(dr * 2))

        self.pos = v2(x, y)

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, pos):
        self._pos = self.sprite.pos = self.refl.pos = self.rect.center = pos

    def update(self):
        self.pos += self.velocity
        self.velocity += GRAVITY

    def collides(self, ano):
        minsep = ano.radius + self.radius
        return self.pos.distance_squared_to(ano.pos) < minsep * minsep



sh = SpatialHash()

for _ in range(BALL_COUNT):
    b = Ball(
        x=random.randint(0, scene.width),
        y=random.randint(0, scene.height),
        radius=random.choice([20, 24, 30])
    )
    sh.insert(b)


def apply_impact(a, b):
    """Resolve the collision between two balls.

    Calculate their closing momentum and apply a fraction of it back as impulse
    to both objects.
    """
    ab = b.pos - a.pos
    try:
        ab.normalize_ip()
    except ValueError:
        ab = v2(1, 0)
    rel_momentum = ab.dot(a.velocity) * a.mass - ab.dot(b.velocity) * b.mass

    if rel_momentum < 0:
        return

    rel_momentum *= ELASTICITY

    a.velocity -= ab * rel_momentum / a.mass
    b.velocity += ab * rel_momentum / b.mass


def separate(a, b, frac=0.5):
    """Move a and b apart.

    Return True if they are now separate.
    """
    ab = a.pos - b.pos
    sep = ab.length()
    overlap = a.radius + b.radius - sep
    if overlap <= 0:
        return

    if sep == 0.0:
        ab = v2(1, 0)
    else:
        ab /= sep
    masses = a.mass + b.mass
    if overlap > 1:
        overlap *= frac

    a.pos += ab * (overlap * b.mass / masses)
    b.pos -= ab * (overlap * a.mass / masses)


def collide_plane(ball, norm, dist, bounce=True):
    overlap = dist - ball.pos.dot(norm) + ball.radius
    if overlap > 0:
        ball.pos += norm * overlap
        if bounce:
            ball.velocity -= norm * norm.dot(ball.velocity) * (1.0 + ELASTICITY)

BOUNDS = [
    (v2(0, -1), -scene.height),
    (v2(1, 0), 0),
    (v2(-1, 0), -scene.width),
]

collisions = set()


def update(dt):
    global collisions
    balls = sh.items
    for b in balls:
        b.update()

    for p in BOUNDS:
        for b in balls:
            collide_plane(b, *p)
    sh.rebuild()

    for p in BOUNDS:
        for b in balls:
            collide_plane(b, *p)

    # Find all collisions occurring this frame
    prev_collisions = collisions
    collisions = set()
    for b in balls:
        possible_collisions = sh.query(b.rect)
        for a in possible_collisions:
            if a is b:
                continue
            if a.collides(b):
                if id(a) > id(b):
                    pair = b, a
                else:
                    pair = a, b

                if pair not in prev_collisions:
                    # We only apply the bounce to the velocity the first time
                    # they collide.
                    apply_impact(*pair)
                collisions.add(pair)

    # Apply several iterations to separate the collisions
    moved = set(b for pair in collisions for b in pair)
    for frac in SEPARATION_STEPS:
        collisions = {pair for pair in collisions if separate(*pair, frac)}

@w2d.event
def on_mouse_move(pos):
    cursor.pos = pos
    r = Rect(0, 0, 30, 30)
    r.center = pos
    possible_collisions = sh.query(r)
    for b in possible_collisions:
        sep = b.pos - pos
        sep.scale_to_length(2)
        b.velocity += sep


@w2d.event
def on_mouse_down():
    w2d.clock.unschedule(update)
    w2d.clock.each_tick(update)

update(0)
w2d.run()



from dataclasses import dataclass

import moderngl

from ..shaders import bind_framebuffer, blend_func
from .base import PostprocessPass


@dataclass
class Bloom:
    """A light bloom effect."""
    ctx: moderngl.Context
    gamma: float = 1.0
    radius: float = 10.0
    intensity: float = 0.5

    camera: 'wasabi2d.scene.Camera' = None
    _blur: PostprocessPass = None
    _fb1: moderngl.Framebuffer = None
    _fb2: moderngl.Framebuffer = None

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._blur = PostprocessPass(
            self.ctx,
            'postprocess/bloom_blur'
        )
        self._copy = PostprocessPass(self.ctx, 'postprocess/copy')

    def draw(self, draw_layer):
        with self.camera.temporary_fb('f1') as fb1:
            with bind_framebuffer(self.ctx, fb1, clear=True):
                draw_layer()

            with self.camera.temporary_fb() as fb2:
                with bind_framebuffer(self.ctx, fb2):
                    with blend_func(self.ctx, moderngl.ONE, moderngl.ZERO):
                        self._blur.render(
                            image=fb1,
                            gamma=self.gamma,
                            blur_direction=(0, 1),
                            radius=self.radius,
                            alpha=1.0
                        )

                self._copy.render(image=fb1)
                with blend_func(self.ctx, moderngl.SRC_ALPHA, moderngl.ONE):
                    self._blur.render(
                        image=fb2,
                        gamma=1e-6,
                        blur_direction=(1, 0),
                        radius=self.radius,
                        alpha=self.intensity
                    )



"""wasabi2d, a 2D game engine using moderngl, pygame and numpy."""
import os

# Don't show Pygame's annoying message because while I might use PyGame,
# I don't appreciate libraries I use communicating with my users.
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

from pygame.math import Vector2
from .game import EventMapper
from .constants import keys, mouse, keymods
from .loaders import sounds
from . import music
from . import clock
from . import tone
from .scene import Scene
from .animation import animate
from .storage import Storage
from .chain import LayerRange
from .primitives.group import Group

event = EventMapper()
del EventMapper

run = event.run

__version__ = (1, 4, 0)
__all__ = [
    'Vector2',
    'event',
    'run',
    'keys', 'mouse', 'keymods',
    'sounds', 'music', 'tone',
    'clock', 'animate',
    'Scene', 'Storage', 'LayerRange',
    'Group'
]



import numpy as np

mapbox_earcut = None

try:
    import mapbox_earcut
except ModuleNotFoundError:
    from ..vendor.earcut import earcut

from ..color import convert_color
from .base import AbstractShape


class Polygon(AbstractShape):
    """An arbitrary polygon."""

    def __init__(
            self,
            layer,
            vertices,
            *,
            pos=(0, 0),
            color=(1, 1, 1, 1),
            stroke_width=1.0):
        self.orig_verts = np.ones((len(vertices), 3), dtype='f4')
        self.orig_verts[:, :2] = vertices

        super().__init__()
        self.layer = layer
        self.pos = pos
        self._stroke_width = stroke_width

        self._color = convert_color(color)
        self._set_dirty()

    @property
    def vertices(self):
        """Get the vertices of the polygon."""
        return self.orig_verts[:, :2].copy()

    @vertices.setter
    def vertices(self, v):
        """Set the vertices of the polygon.

        It is currently not allowed to increase or decrease the number of
        vertices.
        """
        self.orig_verts[:, :2] = v
        self._set_dirty()

    def _stroke_indices(self):
        """Indexes for drawing the stroke as a LINE_STRIP_ADJACENCY."""
        verts = len(self.orig_verts)
        idxs = np.arange(verts, dtype='i4')
        winding = idxs[[-1, *idxs, 0, 1, 2]]
        return winding

    def _fill_indices_mapbox_earcut(self):
        """Indexes for drawing the fill as TRIANGLES.
        
        This version uses the mapbox_earcut library, which is C++, but
        currently doesn't have a Python 3.8 release. See

        https://github.com/skogler/mapbox_earcut_python/issues/2

        """
        verts = self.orig_verts[:, :2]
        rings = np.array([len(verts)], dtype=np.uint32)
        idxs = mapbox_earcut.triangulate_float32(verts, rings)
        return idxs.reshape(-1)

    def _fill_indices_earcut(self):
        """Indexes for drawing the fill as TRIANGLES.
        
        This version uses a pure-Python library, which is vendored into this
        repo.
        """
        verts = self.orig_verts[:, :2].reshape((-1)).astype(np.int64)
        idxs = earcut(verts)
        return np.array(idxs, dtype='i4')

    if mapbox_earcut:
        _fill_indices = _fill_indices_mapbox_earcut
    else:
        _fill_indices = _fill_indices_earcut


class Rect(Polygon):
    """A rectangle."""
    VERTS = np.array([
        (-0.5, -0.5),
        (-0.5, 0.5),
        (0.5, 0.5),
        (0.5, -0.5),
    ], dtype='f4')

    __slots__ = (
        'width', 'height',
    )

    def __init__(
            self,
            layer,
            width,
            height,
            *,
            pos=(0, 0),
            color=(1, 1, 1, 1),
            stroke_width=1):

        self.width = width
        self.height = height

        super().__init__(
            layer,
            self._vertices(),
            pos=pos,
            color=color,
            stroke_width=stroke_width,
        )

    def _vertices(self):
        return np.copy(self.VERTS) * (self.width, self.height)


class PolyLine(Polygon):
    """An unclosed line sequence."""

    def _stroke_indices(self):
        """Indexes for drawing the line as a LINE_STRIP."""
        verts = len(self.orig_verts)
        idxs = np.arange(verts)
        last = verts - 1
        return np.array([0, *idxs, last, last])

    def _fill_indices(self):
        raise NotImplementedError(
            "Lines may not be filled."""
        )



"""Convert colours to greyscale."""
from dataclasses import dataclass

import numpy as np

from .base_matrix import BaseMatrix


SRC_ARR = np.array([
    [0.2126, 0.7152, 0.0722],
] * 3)

DEST_ARR = np.array([
    [0.7874, -0.7152, -0.0722],
    [-0.2126, 0.2848, -0.0722],
    [-0.2126, -0.7152, 0.9278],
])


@dataclass
class Greyscale(BaseMatrix):
    """A sepia effect."""
    amount: float = 1.0

    def get_matrix(self):
        matrix = np.identity(4, dtype=np.float32)
        matrix[:3, :3] = (1 - self.amount) * DEST_ARR
        matrix[:3, :3] += SRC_ARR
        return matrix



"""Example of the light bloom effect."""
from wasabi2d import run, Scene


scene = Scene()

logo = scene.layers[0].add_sprite(
    'wasabi2d',
    pos=(scene.width / 2, scene.height / 2),
)
logo.color = (1.3, 1.3, 1.3, 1)
scene.layers[0].set_effect('bloom', radius=20, intensity=0.25)

run()



"""Color conversion functions."""
from typing import Union, Tuple
import numpy as np
from pygame import Color


def convert_color(c: Union[str, tuple]) -> Tuple[float, float, float, float]:
    """Convert a color to an RGBA tuple."""
    if isinstance(c, str):
        try:
            col = Color(c)
        except ValueError:
            if c.startswith('#'):
                raise ValueError(f"Malformed hex color {c!r}") from None
            raise
        return np.array(memoryview(col), dtype='u1').astype('f4') / 255.0
    else:
        assert 3 <= len(c) <= 4, "Invalid color length"
        return np.array(tuple(c) + (1,) * (4 - len(c)), dtype='f4')


def convert_color_rgb(c: Union[str, tuple]) -> Tuple[float, float, float]:
    """Convert a color to an RGB tuple.

    This accepts the same input formats as convert_color, but raises an
    exception if the alpha value is not 1.

    """
    cc = convert_color(c)
    if abs(cc[3] - 1.0) > 1e-4:
        raise ValueError("Color may not have an alpha component.")
    return cc[:3]



import math
import numpy as np
from dataclasses import dataclass
from typing import Any

import wasabi2d as w2d
from wasabi2d.keyboard import keyboard, keys


scene = w2d.Scene(
    width=1280,
    height=720,
    background="#ccaa88",
)

tilemap = scene.layers[0].add_tile_map()
tilemap.fill_rect(
    'sand_base_1',
    left=0,
    right=scene.width // 64 + 1,
    top=0,
    bottom=scene.height // 64 + 1,
)
tilemap.line(
    'sand_road_lr',
    (2, 1),
    (20, 1)
)
tilemap.line(
    'sand_road_ud',
    (1, 2),
    (1, 8),
)
tilemap.line(
    'sand_road_lr',
    (2, 9),
    (20, 9)
)

tilemap[1, 1] = 'sand_road_dr'
tilemap[1, 4] = 'sand_road_udr'
tilemap.line('sand_road_ld', (2, 4), (6, 8))
tilemap.line('sand_road_ur', (2, 5), (6, 9))
tilemap[1, 9] = 'sand_road_ur'


tilemap.flood_fill(
    ['grass_base_1', 'grass_base_2'],
    (2, 6)
)


scene.layers[1].set_effect('dropshadow', offset=(2, 2))
tank = scene.layers[1].add_sprite('tank_green', pos=(50, 50))
tank.speed = 0


@dataclass
class DrivingController:
    """Control driving an object around."""
    primitive: Any

    #: Acceleration when a key is pressed, in pixels per second
    acceleration: float = 200.0

    #: Fraction of speed retained per frame
    drag: float = 0.1

    #: Rate of turn, in radians/second
    turn: float = 1.5

    #: key bindings
    forward_key: keys = keys.UP
    reverse_key: keys = keys.DOWN
    left_key: keys = keys.LEFT
    right_key: keys = keys.RIGHT

    #: The angle that is "forward" relative to the primitive
    primitive_forward: float = 0.0

    #: The speed of the object. This is maintained by the controller but
    #: can be updated (eg. on collision)
    speed: float = 0.0

    def update(self, dt: float = 1 / 60):
        """Update the primitive."""
        self.speed *= self.drag ** dt
        if keyboard.up:
            self.speed += self.acceleration * dt
        elif keyboard.down:
            self.speed -= self.acceleration * dt

        if keyboard.left:
            self.primitive.angle -= np.copysign(self.turn * dt, self.speed)
        elif keyboard.right:
            self.primitive.angle += np.copysign(self.turn * dt, self.speed)

        displacement = self.speed * dt
        self.primitive.pos += displacement * self.forward_vector()

    __basis = np.array([np.pi / 2, 0], dtype=np.float32)

    def forward_vector(self) -> np.ndarray:
        """Get a unit vector in the forward direction."""
        return np.sin(
            self.__basis + (self.primitive.angle + self.primitive_forward)
        )


tank_control = DrivingController(
    tank,
    acceleration=500,
    primitive_forward=math.pi / 2
)


@w2d.event
def update(dt):
    tank_control.update(dt)
    scene.camera.pos = tank.pos


w2d.run()



"""An effect where we keep trails from previous frames."""
from dataclasses import dataclass

import moderngl

from ..clock import Clock, default_clock
from ..shaders import bind_framebuffer, blend_func
from .base import PostprocessPass


@dataclass
class Trails:
    """A trails effect."""
    ctx: moderngl.Context
    fade: float = 0.9
    alpha: float = 1.0
    clock: Clock = default_clock

    camera: 'wasabi2d.scene.Camera' = None
    _trails_buf: moderngl.Framebuffer = None

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._trails_buf = camera._make_fb('f4')
        self._fade_pass = PostprocessPass(
            self.ctx,
            'postprocess/trails_fade',
            send_uvs=False
        )
        self._composite_pass = PostprocessPass(
            self.ctx,
            'postprocess/trails_composite',
        )
        self.t = self.clock.t

    def __del__(self):
        if self._trails_buf:
            self._trails_buf.release()

    def draw(self, draw_layer):
        dt = self.clock.t - self.t
        self.t = self.clock.t

        with self.camera.temporary_fbs(1, 'f2') as (tmp,):
            with blend_func(self.ctx, 'a', '1-a', '1', '1-a'):
                with bind_framebuffer(self.ctx, tmp, clear=True):
                    draw_layer()

                self._composite_pass.render(
                    fb=self._trails_buf,
                    alpha=self.alpha,
                )
                self._composite_pass.render(
                    fb=tmp,
                    alpha=1.0
                )

            with bind_framebuffer(self.ctx, self._trails_buf):
                with blend_func(self.ctx, 'a', '1-a', '1', '1-a'):
                    self._composite_pass.render(
                        fb=tmp,
                        alpha=1.0
                    )
                with blend_func(self.ctx, 0, 1, 0, 'a'):
                    self._fade_pass.render(
                        fade=self.fade ** dt
                    )



"""Classes for representing the render graph."""
from typing import Optional, List, Tuple
from dataclasses import dataclass
from functools import partial
import importlib
from collections import Counter
from contextlib import contextmanager

import numpy as np

from .color import convert_color


class ChainNode:
    """Mix-in class for chain nodes to wrap them in an additional effect."""

    def wrap_effect(self, name: str, **parameters) -> 'Effect':
        """Wrap this layer range in an effect.

        The name and parameters are identical to ``Layer.set_effect()``.

        This is a shortcut for constructing the effect object directly.

        :param name: The name of an effect to enable.
        :param parameters: Parameters to control the effect.
        :return: An Effect object, which can be used to change parameters for
                 the effect, or to wrap this effect in another effect.

        """
        return Effect(self, name, parameters)


@dataclass
class LayerRange(ChainNode):
    """Render a range of layers, in order of lowest id to highest id.

    Only existing layers are rendered; any that do not are skipped - but will
    be rendered as soon as they are created.

    Unlike range(), start and stop are *inclusive*.
    """
    start: Optional[int] = None
    stop: Optional[int] = None

    def draw(self, scene):
        """Draw the selected layers."""
        layers = scene.layers

        start = self.start if self.start is not None else -np.inf
        stop = self.stop if self.stop is not None else np.inf

        for k in sorted(layers):
            if start <= k <= stop:
                layers[k]._draw()


@dataclass
class Layers(ChainNode):
    """Render specific layers in the specified order.

    Only existing layers are rendered; any that do not are skipped - but will
    be rendered as soon as they are created.

    """
    layers: List[int]

    def draw(self, scene):
        for layer_num in self.layers:
            if layer_num in scene.layers:
                scene.layers[layer_num]._draw()


def to_node(val):
    """Convert a user-provided value to a ChainNode.

    This allows several shortcuts for specifying layers to a node.
    """
    if isinstance(val, ChainNode):
        return val

    if isinstance(val, int):
        return Layers([val])

    if isinstance(val, list):
        types = Counter(map(type, val))

        if len(types) == 1:
            typ, _ = types.popitem()
            if typ is int:
                return Layers(val)

        return Merge([to_node(v) for v in val])

    raise TypeError(f"Cannot convert {val!r} to ChainNode.")


@dataclass
class Merge(ChainNode):
    """Draw each of the given chain nodes in order."""
    nodes: List[ChainNode]

    def draw(self, scene):
        for node in self.nodes:
            node.draw(scene)


class Effect(ChainNode):
    """Apply a post-processing effect to the contained subchain."""
    __slots__ = (
        'draw', '_effect', '_subchain', '_camera', '_effect_keys',
    )

    def __init__(self, child_node: ChainNode, effect: str, params: dict):
        child_node = to_node(child_node)

        mod = importlib.import_module(f'wasabi2d.effects.{effect}')
        cls = getattr(mod, effect.title())
        self._effect_keys = set(cls.__dataclass_fields__)

        unexpected = params.keys() - self._effect_keys
        if unexpected:
            raise TypeError(
                f"{effect} does not accept attributes {unexpected}"
            )

        self._effect = None
        self._subchain = child_node
        self._camera = None

        def draw(scene):
            """Bind the context and instantiate the effect."""
            ctx = scene.ctx
            self._effect = cls(ctx, **params)
            self.draw = self.real_draw
            self.draw(scene)
        self.draw = draw

    def real_draw(self, scene):
        if scene.camera is not self._camera:
            self._camera = scene.camera
            self._effect._set_camera(self._camera)
        self._effect.draw(partial(self._subchain.draw, scene))

    def __getattr__(self, k):
        if k in self._effect_keys:
            return getattr(self._effect, k)
        return object.__getattr__(self, k)

    def __setattr__(self, k, v):
        if not k.startswith('_') and k in self._effect_keys:
            setattr(self._effect, k, v)
        object.__setattr__(self, k, v)


MASK_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;
uniform sampler2D paint;
uniform sampler2D mask;

void main()
{
    vec4 paint_frag = texture(paint, uv);
    vec4 mask = texture(mask, uv);

    float a = %s;

    if (paint_frag.a * a < 1e-6) {
        discard;
    }
    f_color = vec4(paint_frag.rgb / paint_frag.a, paint_frag.a * a);
}
"""

MASK_FUNCS = {
    'inside': 'mask.a',
    'outside': '1 - mask.a',
    'luminance': 'dot(mask.rgb, vec3(0.3, 0.6, 0.1))',
}


@contextmanager
def rendered_node(scene, node):
    """Context manager to render a node to a temporary framebuffer.

    The framebuffer is reserved for the duration of the context, but then
    released.

    """
    camera = scene.camera
    with camera.temporary_fb() as fb:
        with camera.bind_framebuffer(fb):
            node.draw(scene)
        yield fb


@dataclass
class Mask(ChainNode):
    """Draw one 'paint' layer only where the 'mask' layer is opaque."""

    mask: ChainNode
    paint: ChainNode
    function: str = 'inside'

    def __post_init__(self):
        self.mask = to_node(self.mask)
        self.paint = to_node(self.paint)

    def draw(self, scene):
        """Draw the effect."""
        with rendered_node(scene, self.mask) as mask, \
                rendered_node(scene, self.paint) as paint:
            scene.camera.run_shader(
                MASK_PROG % MASK_FUNCS[self.function],
                paint=paint,
                mask=mask,
            )


class Fill(ChainNode):
    """Fill the screen with a single colour.

    This can be useful as input to another chain node, such as a mask. Note
    that ``Scene.background`` is also available if you simply want to clear
    the screen to a certain colour before rendering.

    """
    color: Tuple[float, float, float, float]

    def __init__(self, color):
        self.color = color

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, v):
        self._color = convert_color(v)

    def draw(self, scene):
        """Draw the effect."""
        scene.ctx.clear(*self._color)


DISPLACEMENT_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;
uniform sampler2D paint;
uniform sampler2D displacement;
uniform float scale;


void main()
{
    vec4 disp_frag = texture(displacement, uv);

    if (disp_frag.a < 1e-6) {
        discard;
    }
    vec3 unpremultiplied = disp_frag.rgb / disp_frag.a;
    float r = unpremultiplied.r;
    float g = unpremultiplied.g;
    float b = unpremultiplied.b;
    vec2 offset = (vec2(%s, %s) - vec2(0.5, 0.5)) * (2.0 * scale);

    vec4 paint_frag = texture(paint, uv + offset / textureSize(paint, 0));
    if (paint_frag.a < 1e-6) {
        discard;
    }
    f_color = vec4(paint_frag.rgb / paint_frag.a, paint_frag.a * disp_frag.a);
}
"""


@dataclass
class DisplacementMap(ChainNode):
    """Offset one input using the values of another."""

    paint: ChainNode
    displacement: ChainNode
    scale: float = 10.0
    x_channel: str = 'r'
    y_channel: str = 'g'

    def draw(self, scene):
        """Draw the effect."""
        with rendered_node(scene, self.displacement) as disp, \
                rendered_node(scene, self.paint) as paint:
            scene.camera.run_shader(
                DISPLACEMENT_PROG % (self.x_channel, self.y_channel),
                paint=paint,
                displacement=disp,
                scale=self.scale
            )



import math
from functools import partial

import numpy as np
import moderngl

from ..color import convert_color
from ..allocators.vertlists import VAO
from .polygons import AbstractShape
from ..rect import ZRect
from ..descriptors import CallbackProp
from ..effects.base import PostprocessPass
from ..shaders import bind_framebuffer, blend_func


class PolyVAO(VAO):
    """VAO object that renders with multisampling."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.composite_prog = PostprocessPass(
            self.ctx,
            'postprocess/multisample_blend'
        )

    def render(self, camera):
        samples = min(self.ctx.max_samples, 4)
        with camera.temporary_fb(samples=samples) as fb:
            with bind_framebuffer(self.ctx, fb, clear=True):
                super().render(camera)
            self.composite_prog.render(image=fb, samples=samples)


def color_vao(
        mode: int,
        ctx: moderngl.Context,
        shadermgr: 'wasabi2d.layers.ShaderManager') -> VAO:
    """Build a BAO for rendering plain colored objects."""
    return PolyVAO(
        mode=mode,
        ctx=ctx,
        prog=shadermgr.load('primitives/flat_color'),
        dtype=np.dtype([
            ('in_vert', '2f4'),
            ('in_color', '4f4'),
        ])
    )


def line_vao(
        ctx: moderngl.Context,
        shadermgr: 'wasabi2d.layers.ShaderManager') -> VAO:
    """Build a VAO for rendering lines."""
    return PolyVAO(
        mode=moderngl.LINE_STRIP_ADJACENCY,
        ctx=ctx,
        prog=shadermgr.load(
            'primitives/wide_line',
            'primitives/wide_line',
            'primitives/flat_color',
        ),
        dtype=np.dtype([
            ('in_vert', '2f4'),
            ('in_color', '4f4'),
            ('in_linewidth', 'f4'),
        ])
    )


#: Construct a VAO for rendering the fill of shapes
shape_vao = partial(color_vao, moderngl.TRIANGLES)


class Circle(AbstractShape):
    """A circle drawn with lines."""

    __slots__ = (
        'layer', 'segments',
        '_radius',
        'orig_verts',
    )

    def __init__(
            self,
            layer,
            pos=(0, 0),
            radius=100,
            color=(1, 1, 1, 1),
            stroke_width=1.0,
            segments=None):
        super().__init__()
        self.layer = layer
        self.segments = segments or max(4, round(math.pi * radius))
        self.pos = pos
        self._stroke_width = stroke_width

        # Generate verts now
        theta = np.linspace(0, 2 * np.pi, self.segments).reshape((-1, 1))
        self.base_verts = np.hstack([
            np.cos(theta),
            np.sin(theta),
        ]).astype('f4')

        # There's a duplicate vertex so move the first vertex to the center
        self.base_verts[0][:2] = 0

        # Placeholder for scaled verts
        self.orig_verts = np.ones((self.segments, 3), dtype='f4')

        # Assigning radius generates self.orig_verts
        self.radius = radius

        self._color = convert_color(color)
        self._set_dirty()

    def _on_set_radius(self):
        np.multiply(
            self.base_verts,
            self._radius,
            self.orig_verts[:, :2]
        )
        self._set_dirty()

    radius = CallbackProp(_on_set_radius)

    def _stroke_indices(self):
        """Indexes for drawing the stroke as a LINE_STRIP."""
        n = self.segments
        return np.array(
            [n - 1, *range(1, n), 1, 2],
            dtype='i4'
        )

    def _fill_indices(self):
        """Indexes for drawing the fill as TRIANGLES."""
        idxs = np.array([
            (0, i, i + 1)
            for i in range(1, self.segments)
        ], dtype='i4')
        idxs[-1][2] = 1
        return idxs.reshape((-1))

    @property
    def bounds(self):
        x, y = self.pos
        r = self.radius * self.scale
        return ZRect(
            (x - r, y - r),
            (r * 2, r * 2),
        )



"""Base class for effects that simply apply a color matrix."""
from dataclasses import dataclass
import abc

import moderngl
import numpy as np

from ..shaders import bind_framebuffer
from .base import PostprocessPass


@dataclass
class BaseMatrix(metaclass=abc.ABCMeta):
    ctx: moderngl.Context

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera

        self._color_mat = PostprocessPass(
            self.ctx,
            'postprocess/color_matrix'
        )

    @abc.abstractmethod
    def get_matrix(self) -> np.ndarray:
        """Get the matrix to draw.

        The matrix should be (4x4).
        """

    def draw(self, draw_layer):
        """Subclasses should implement this to pass the matrix to draw."""
        with self.camera.temporary_fb() as fb:
            with bind_framebuffer(self.ctx, fb, clear=True):
                draw_layer()

            self._color_mat.render(
                image=fb,
                color_matrix=self.get_matrix(),
            )



import math
import random
import wasabi2d as w2d
from wasabi2d import clock, Vector2, animate


scene = w2d.Scene()
trail = scene.layers[-1].add_particle_group(max_age=2, grow=0.1)
ship = scene.layers[0].add_sprite(
    'ship',
    pos=(scene.width / 2, scene.height / 2)
)


async def drive_ship():
    while True:
        x, y = ship.pos
        tx = random.uniform(50, scene.width - 50)
        ty = random.uniform(50, scene.height - 50)

        dx = tx - x
        dy = ty - y
        angle = math.atan2(dy, dx)
        dist = math.hypot(dx, dy)
        duration = (dist / 500) ** 0.5

        # Rotate to face
        await animate(ship, duration=0.5, angle=angle)

        # Begin emitting particles
        clock.coro.run(thrust(duration * 0.5))

        # Move
        scene.layers[0].set_effect('trails', fade=1e-2)
        await animate(
            ship,
            duration=duration,
            tween='accel_decel',
            pos=(tx, ty),
        )
        scene.layers[0].clear_effect()


async def thrust(duration):
    """Fire a little burst of thrust."""
    prev_pos = Vector2(*ship.pos)
    async for _ in clock.coro.frames(seconds=duration):
        v = Vector2(*ship.pos) - prev_pos
        num = v.length() // 50
        if num:
            trail.emit(
                num,
                pos=ship.pos,
                vel=-v.normalize() * 100,
                vel_spread=10,
                size=3,
                color='#80ffff',
            )


clock.coro.run(drive_ship())
w2d.run()



"""Separable Gaussian blur."""
from dataclasses import dataclass

import moderngl

from ..shaders import bind_framebuffer
from .base import PostprocessPass


@dataclass
class Blur:
    """A light bloom effect."""
    ctx: moderngl.Context
    radius: float = 10.0

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._outer_fb = self.ctx.screen
        self._blur = PostprocessPass(
            self.ctx,
            'postprocess/blur',
        )

    def draw(self, draw_layer):
        with self.camera.temporary_fbs(2, 'f2') as (fb1, fb2):
            with bind_framebuffer(self.ctx, fb1, clear=True):
                draw_layer()

            # Hold onto this for the drop shadow effect to use.
            # This is very much cheating.
            self._fb1 = fb1

            with bind_framebuffer(self.ctx, fb2, clear=True):
                self._blur.render(
                    image=fb1,
                    blur_direction=(0, 1),
                    radius=self.radius,
                )

            self._blur.render(
                image=fb2,
                blur_direction=(1, 0),
                radius=self.radius,
            )



"""A distortion of the scene inwards/outwards."""
from dataclasses import dataclass

import moderngl

from ..shaders import bind_framebuffer
from .base import PostprocessPass


@dataclass
class Punch:
    """A distortion of the scene inwards/outwards."""
    ctx: moderngl.Context
    factor: float = 0.9

    camera: 'wasabi2d.scene.Camera' = None
    _pass: PostprocessPass = None
    _fb: moderngl.Framebuffer = None

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._pass = PostprocessPass(self.ctx, 'postprocess/punch')

    def draw(self, draw_layer):
        with self.camera.temporary_fbs(1, 'f2') as (fb,):
            with bind_framebuffer(self.ctx, fb, clear=True):
                draw_layer()

            self._pass.render(
                fb=fb,
                factor=self.factor,
            )



"""Sparse tile maps."""
import random
from itertools import product
from typing import Tuple, Dict, Set, Optional, List, TypeVar, Union
from collections import deque

import moderngl
import numpy as np
import bresenham

import wasabi2d
from ..allocators.abstract import FreeListAllocator, NoCapacity
from ..allocators.vertlists import dtype_to_moderngl


def grow(arr: np.ndarray, newsize: int) -> np.ndarray:
    shape = list(arr.shape)
    shape[0] = newsize

    new = np.empty_like(arr, shape=shape)
    new[:len(arr)] = arr
    return new


T = TypeVar('T')


class TileManager:
    """Manage tile blocks in a texture and a vertex buffer enumerating them.

    Each vertex buffer entry will contain the block coordinates for the block
    and the location in the texture where the texture data is found. This
    latter location is stored as an integer index.

    Each texture block will be a 64x64 region of uint8 tile values.

    """
    ctx: moderngl.Context

    # Allocator for texture areas
    alloc: FreeListAllocator
    texture: moderngl.Texture

    # Vertices and a backing memory array
    verts: moderngl.Buffer
    vertdata: np.ndarray
    verts_dirty: bool = True

    block_map: Dict[Tuple[int, int], int]
    texture_blocks: Dict[int, np.ndarray]
    dirty_blocks: Set[int]

    NP_DTYPE = np.dtype([
        ('in_vert', '2i4'),
        ('in_tilemap_block', 'u2'),
    ])
    MGL_DTYPE = dtype_to_moderngl(NP_DTYPE)

    def __init__(self, ctx: moderngl.Context):
        self.ctx = ctx

        # TODO: this allocator is overkill given that we cannot actually
        # free tile blocks individually; replace with a simple counter
        self.alloc = FreeListAllocator()
        self.block_map = {}
        self.dirty_blocks = set()
        self.texture_blocks = {}

        self.mktex(self.alloc.capacity)
        self.vertdata = np.empty(self.alloc.capacity, dtype=self.NP_DTYPE)
        self.verts = self.ctx.buffer(self.vertdata)

    def mktex(self, capacity: int):
        """Create a new texture."""
        self.texture = self.ctx.texture(
            (capacity * 64, 64), 1,
            dtype='f1'
        )

    def mktile(self) -> np.ndarray:
        """Create a tile in memory."""
        return np.zeros((64, 64), dtype=np.uint8)

    def _resize(self, new_capacity: int):
        """Resize to hold new_capacity tiles."""
        self.texture.release()
        self.mktex(new_capacity)
        self.alloc.grow(new_capacity)
        self.dirty_blocks.update(self.block_map.values())
        self.vertdata = grow(self.vertdata, new_capacity)
        self.verts.orphan(new_capacity)
        self.verts_dirty = True

    def release(self):
        """Release the ModernGL objects."""
        if self.verts:
            self.verts.release()
            self.verts = None
        if self.texture:
            self.texture.release()
            self.texture = None

    def __del__(self):
        self.release()

    def create_block(self, pos: Tuple[int, int]) -> np.ndarray:
        try:
            tile_id = self.alloc.alloc()
        except NoCapacity as e:
            self.resize(e.capacity)
            tile_id = self.alloc.alloc()

        vert_id = len(self.block_map)

        tile = self.mktile()
        self.block_map[pos] = tile_id
        self.texture_blocks[tile_id] = tile
        self.dirty_blocks.add(tile_id)

        self.vertdata[vert_id] = pos, tile_id

        stride = self.vertdata.strides[0]
        self.verts.write(
            data=self.vertdata[vert_id:vert_id + 1],
            offset=stride * vert_id
        )
        return tile

    def clear(self):
        """Clear all allocations."""
        self.alloc = FreeListAllocator(self.alloc.capacity)
        self.block_map.clear()
        self.texture_blocks.clear()
        self.dirty_blocks.clear()

    def __len__(self):
        return len(self.block_map)

    def get_or_create_block(self, pos: Tuple[int, int]) -> np.ndarray:
        """Get or create the block for the given coordinates.

        This method is intended for updates to the block and will mark the
        block as dirty.

        """
        id = self.block_map.get(pos)
        if id is None:
            return self.create_block(pos)
        self.dirty_blocks.add(id)
        return self.texture_blocks[id]

    def get_block(self, pos: Tuple[int, int]) -> Optional[np.ndarray]:
        """Retrieve the block with the given coordinates if it exists.

        This method does not mark the block as dirty.
        """
        id = self.block_map.get(pos)
        if id is None:
            return None
        return self.texture_blocks[id]

    def touch_block(self, pos: Tuple[int, int]):
        """Mark a block as dirty."""
        self.dirty_blocks.append(self.block_map[pos])

    def bind_texture(self, unit: int):
        """Bind the texture to a texture unit."""
        for tile_id in self.dirty_blocks:
            tile = self.texture_blocks[tile_id]
            self.texture.write(
                tile,
                (tile_id * 64, 0, 64, 64),
                alignment=1
            )
        self.dirty_blocks.clear()
        self.texture.use(unit)


class TileMap:
    """A sparse tile map."""
    _tiles: List[str]
    layer: 'wasabi2d.layers.Layer'
    _tilemgr: TileManager

    # A texture holding texture coordinates for tiles in the map,
    # and its backing store in memory.
    _tile_tex: moderngl.Texture = None
    _texdata: np.ndarray
    _tile_tex_dirty: bool = False

    vao: moderngl.VertexArray = None

    size: Optional[Tuple[int, int]] = None
    tex: Optional = None
    block_size: Optional[Tuple[int, int]] = None

    def __init__(self, layer, *, tile_size=None, any_size_tile=False):
        super().__init__()
        self.layer = layer
        self.atlas = self.layer.group.atlas
        self._tiles = [None]

        if tile_size:
            w, h = self.size = tile_size
            self.block_size = w * 64, h * 64
        if any_size_tile and not tile_size:
            raise ValueError(
                "tile_size must be given to use any_size_tile"
            )
        self.any_size_tile = any_size_tile

        self._texdata = np.zeros((256, 4, 2), dtype=np.float32)
        self._names = {}
        self._tile_tex = self.layer.ctx.texture(
            (2, 256),
            4,
            dtype='f4'
        )
        self._tile_tex.filter = moderngl.NEAREST, moderngl.NEAREST

        self._tilemgr = TileManager(self.layer.ctx)

        shadermgr = layer.group.shadermgr

        self.prog = shadermgr.load('primitives/tile_map')
        self.vao = self.layer.ctx.vertex_array(
            self.prog,
            [(self._tilemgr.verts, *TileManager.MGL_DTYPE)]
        )
        layer.arrays[id(self)] = self

    def release(self):
        """Release OpenGL resources associated with this TileMap."""
        if self.vao:
            self.vao.release()
        if self._tile_tex:
            self._tile_tex.release()
        if self._tilemgr:
            self._tilemgr.release()
        self._tilemgr = self._tile_tex = self.vao = None

    def __del__(self):
        self.release()

    def _map_name(self, name: str) -> int:
        """Given an image name, return an integer mapping for it.

        If the image is not already in the tile list, assign it a new ID.
        """
        if name is None:
            return 0
        try:
            return self._names[name]
        except KeyError:
            pass

        id = len(self._tiles)

        region = self.atlas.get(name)
        rsize = region.width, region.height
        if id >= 256:
            raise ValueError(
                f"Cannot insert {name} to {self!r}; "
                "{len(self._tiles)} tile slots are already in use."
            )
        if id > 1:
            if rsize != self.size and not self.any_size_tile:
                raise ValueError(
                    f"Size of {name} ({rsize}) does not match "
                    f"previous tiles {self.size}"
                )
            if region.tex is not self.tex:
                raise ValueError(
                    f"Tile size is too large {rsize}"
                )
        else:
            self.tex = region.tex
            if not self.size:
                self.size = rsize
                self.block_size = region.width * 64, region.height * 64

        tl, tr, br, bl = region.texcoords.astype(np.float32)

        self._texdata[id, :] = [tl, tr - tl, bl - tl, (0, 0)]
        self._tile_tex_dirty = True
        self._tiles.append(name)
        self._names[name] = id
        return id

    def _value_gen(self, value):
        """Return a generator for tile values."""
        if value is None or isinstance(value, str):
            id = self._map_name(value)
            while True:
                yield id

        choices = [self._map_name(v) for v in value]
        while True:
            yield random.choice(choices)

    def fill_rect(
        self,
        value: Union[str, List[str], None],
        left: int,
        top: int,
        right: int,
        bottom: int,
    ):
        """Fill a rectangle of the tile map.

        `value` can be a string to fill with just one tile value or a list of
        strings in order to fill with one of a range of choices. You can also
        pass ``None`` in order to clear tiles instead of setting them.

        Note that right/bottom are exclusive.
        """
        cells = product(range(left, right), range(top, bottom))
        for pos, v in zip(cells, self._value_gen(value)):
            self._set(pos, v)

    def line(
        self,
        value: Union[str, List[str], None],
        start: Tuple[int, int],
        stop: Tuple[int, int],
    ):
        """Fill a line from coordinates start to stop.

        `value` can be a string to fill with just one tile value or a list of
        strings in order to fill with one of a range of choices. You can also
        pass ``None`` in order to clear tiles instead of setting them.

        """
        cells = bresenham.bresenham(*start, *stop)
        for pos, v in zip(cells, self._value_gen(value)):
            self._set(pos, v)

    def flood_fill(
        self,
        value: Union[str, List[str], None],
        start: Tuple[int, int],
        *,
        limit: int = 10_000
    ):
        """Flood fill from the given position.

        `value` can be a string to fill with just one tile value or a list of
        strings in order to fill with one of a range of choices. You can also
        pass ``None`` in order to clear tiles instead of setting them.

        Because the tile map is unbounded, `limit` caps the number of
        tiles that can be considered for filling. If this limit is hit
        then an exception is raised and no tiles are updated.

        """
        values = self._value_gen(value)
        match_value = self._get(start)
        queue = deque([start])
        seen = {start, }
        fill = []
        for _ in range(limit):
            if not queue:
                break
            pos = queue.popleft()
            if self._get(pos) != match_value:
                continue
            fill.append(pos)

            x, y = pos
            neighbours = {
                (x + 1, y),
                (x - 1, y),
                (x, y + 1),
                (x, y - 1),
            } - seen
            queue.extend(neighbours)
            seen.update(neighbours)
        else:
            raise ValueError(f"Hit flood-fill limit of {limit} tiles")
        for pos, v in zip(fill, values):
            self._set(pos, v)

    def __getitem__(self, pos: Tuple[int, int]) -> str:
        """Get the tile at the given position."""
        v = self.get(pos)
        if not v:
            raise KeyError(f"No tile at position {pos}")
        return v

    def get(self, pos: Tuple[int, int], default: T = None) -> Union[str, T]:
        """Get the tile at the given position.

        If there is no tile at that position, `default` is returned.
        """
        id = self._get(pos)
        if id == 0:
            return default
        return self._tiles[id]

    def _get(self, pos: Tuple[int, int]) -> int:
        """Get the tile id at the given position.

        If there is no tile at that position, 0 is returned.
        """
        cell, (x, y) = np.divmod(pos, 64)
        block = self._tilemgr.get_block(tuple(cell))
        return block is not None and block[y, x] or 0

    def __setitem__(self, pos: Tuple[int, int], value: str):
        """Set the tile at the given position."""
        id = self._map_name(value)
        self._set(pos, id)

    def _set(self, pos, tid: int):
        """Set a tile id at a given location."""
        cell, (x, y) = np.divmod(pos, 64)
        block = self._tilemgr.get_or_create_block(tuple(cell))
        block[y, x] = tid

    def setdefault(self, pos: Tuple[int, int], value: str) -> str:
        """Set a tile in the tile map if it is not set.

        Return the tile that is set in this cell after the call.
        """
        cell, (x, y) = np.divmod(pos, 64)
        block = self._tilemgr.get_or_create_block(tuple(cell))
        idx = y, x
        v = block[idx] = block[idx] or self._map_name(value)
        return self._tiles[v]

    def __delitem__(self, pos: Tuple[int, int]):
        """Clear the tile at the given position."""
        self._set(pos, 0)

    def clear(self):
        """Clear the tile map."""
        self._tilemgr.clear()
        self._tiles.clear()
        self._names.clear()

    @property
    def bounds(self):
        """todo"""
        raise NotImplementedError()

    def delete(self):
        """Delete this primitive."""
        self.layer._dirty.discard(self)
        self.layer.objects.discard(self)
        self.layer = None
        self.release()

    def render(self, camera: wasabi2d.scene.Camera):
        blocks = len(self._tilemgr.block_map)
        if not blocks:
            return
        self._tilemgr.bind_texture(0)
        self.prog['tiles'] = 0

        if self._tile_tex_dirty:
            self._tile_tex.write(self._texdata)
            self._tile_tex_dirty = False

        self._tile_tex.use(1)
        self.prog['tilemap_coords'] = 1

        self.tex.tex.use(2)
        self.prog['tex'] = 2
        self.prog['block_size'] = self.block_size
        self.prog['screen_width'] = camera.width

        self.vao.render(
            mode=moderngl.POINTS,
            vertices=blocks
        )



import random
import wasabi2d as w2d
from wasabi2d import clock, Vector2, animate


scene = w2d.Scene()
scene.background = '#ccddff'


async def explode(pos):
    """Create an explosion at pos."""
    scene.camera.screen_shake()
    sprite = scene.layers[1].add_sprite('explosion', pos=pos)
    sprite.color = (1, 1, 1, 0)

    animate(
        sprite,
        duration=0.9,
        angle=10,
    )

    # Explode phase
    await animate(
        sprite,
        duration=0.3,
        tween='accelerate',
        scale=3,
        color=(1, 1, 1, 1),
    )

    # Twist phase
    await clock.coro.sleep(0.1)

    # Collapse phase
    await animate(
        sprite,
        duration=0.5,
        scale=1,
        color=(0, 0, 0, 0),
    )

    # Delete it again
    sprite.delete()


async def spawn_explosions():
    """Continue spawning explosions."""
    while True:
        px = random.uniform(50, scene.width - 50)
        py = random.uniform(50, scene.height - 50)
        clock.coro.run(explode((px, py)))
        await clock.coro.sleep(random.uniform(0.5, 3))


clock.coro.run(spawn_explosions())
w2d.run()



from wasabi2d import Scene, clock, run


scene = Scene()


x = scene.width / 2
labels = [
    scene.layers[0].add_label("Hello", pos=(x, 200)),
    scene.layers[0].add_label("World", pos=(x, 300)),
    scene.layers[0].add_label("", pos=(x, 400)),
]


def rotate():
    """Rotate the text strings between the labels."""
    texts = [l.text for l in labels]
    texts.insert(0, texts.pop())
    for l, t in zip(labels, texts):
        l.text = t


clock.schedule_interval(rotate, 0.4)

run()



"""Example of the sepia effect."""
import wasabi2d as w2d

scene = w2d.Scene()

photo = scene.layers[0].add_sprite(
    'positano',
    pos=(scene.width / 2, scene.height / 2),
)
photo.scale = max(
    scene.width / photo.width,
    scene.height / photo.height
)
effect = scene.layers[0].set_effect('posterize')

levels_label = scene.layers[1].add_label(
    f'Levels: {effect.levels}',
    align='center',
    pos=(scene.width / 2, scene.height - 40),
    color='white',
    fontsize=20,
)
gamma_label = scene.layers[1].add_label(
    f'Gamma: {effect.gamma:0.2f}',
    align='center',
    pos=(scene.width / 2, scene.height - 20),
    color='white',
    fontsize=20,
)
scene.layers[1].set_effect('dropshadow', radius=2)


@w2d.event
def on_mouse_move(pos):
    x, y = pos
    fracx = x / scene.width
    fracy = y / scene.height
    effect.levels = round(18 * fracx) + 1
    effect.gamma = 0.2 + 2 * fracy
    levels_label.text = f'Levels: {effect.levels}'
    gamma_label.text = f'Gamma: {effect.gamma:0.2f}'


w2d.run()



"""Clock/event scheduler.

This is a Pygame implementation of a scheduler inspired by the clock
classes in Pyglet.

This clock holds weak references to callbacks by default. This is useful to
avoid accidentally "leaking" objects; with strong references it is very easy
to accidentally create a situation where a leaked object is perpetually being
updated even though nothing else references it. Weak references ensure that
objects are only updated if they are referenced elsewhere.

"""
import heapq
import warnings
from weakref import ref
from itertools import chain, count
from functools import total_ordering
from collections import namedtuple
from types import MethodType
import types


__all__ = [
    'Clock', 'schedule', 'schedule_interval', 'unschedule',
    'schedule_unique', 'each_tick', 'call_soon', 'coro',
]

# This type can't be weakreffed in Python 3.4
builtin_function_or_method = type(open)


def weak_method(method):
    """Quick weak method ref in case users aren't using Python 3.4"""
    selfref = ref(method.__self__)
    funcref = ref(method.__func__)

    def weakref():
        self = selfref()
        func = funcref()
        if self is None or func is None:
            return None
        return func.__get__(self)
    return weakref


def mkref(o):
    if isinstance(o, MethodType):
        return weak_method(o)
    else:
        try:
            return ref(o)
        except TypeError:
            if isinstance(o, builtin_function_or_method):
                return lambda: o
            raise


@total_ordering
class Event:
    """An event scheduled for a future time.

    Events are ordered by their scheduled execution time.

    """
    def __init__(self, time, cb, strong=False, repeat=None):
        self.time = time
        self.repeat = repeat
        self.cb = mkref(cb) if not strong else lambda: cb
        self.name = str(cb)
        self.repeat = repeat

    def __lt__(self, ano):
        return self.time < ano.time

    def __eq__(self, ano):
        return self.time == ano.time

    @property
    def callback(self):
        return self.cb()


WaitDelay = namedtuple('Delay', 'seconds')
WaitTick = object()


class Future:
    """An object that can be awaited.

    When awaited, yield a value that indicates the event to wait for.
    """
    __slots__ = ('val', 'awaited')

    def __init__(self, val):
        self.val = val
        self.awaited = False

    def __await__(self):
        self.awaited = True
        yield self

    def __del__(self):
        if not self.awaited:
            warnings.warn(ResourceWarning("wasabi2d future was not awaited"))


class WaitEvent:
    """Await some condition that will arise in future."""
    def __init__(self):
        self.done = False
        self._result = None
        self._on_ready = None

    def when_ready(self, callback):
        self._on_ready = callback
        if self.done:
            callback()

    def set(self, result):
        """Set the result for the event."""
        self.done = True
        self._result = result
        self._on_ready()

    def get_result(self):
        """Get the result."""
        if not self.done:
            raise ValueError("Event is not complete.")
        return self._result


class Coroutines:
    """Namespace for coroutine operations on a clock."""

    class Cancelled(Exception):
        """Raised inside a coroutine when a task is cancelled."""

    def __init__(self, clock):
        self.clock = clock
        self._ready_events = set()

    def _delay(self, seconds):
        """Get a future for a delay."""
        return Future(WaitDelay(seconds))

    def _frame(self):
        """Get a future for the next frame."""
        return Future(WaitTick)

    def _event(self):
        return Future(WaitEvent())

    async def sleep(self, seconds):
        """Sleep for the given time in seconds."""
        await self._delay(seconds)
        return seconds

    async def next_frame(self):
        """Await the next frame. Return the time elapsed."""
        start = self.clock.t
        await self._frame()
        return self.clock.t - start

    async def frames(self, *, seconds=None, frames=None):
        """Iterate over multiple frames, yielding the total time.

        For example::

            async for t in clock.coro.frames(seconds=10):
                percent = t * 10.0
                print(f"Waiting {percent}%")

        If seconds or frames are given these are the limit on the duration of
        the loop; otherwise iterate forever.

        If limiting by seconds, you are guaranteed to receive an event after
        exactly ``seconds``, regardless of frame rate, in order to ensure that
        any effect is complete.

        """
        if seconds is not None and frames is not None:
            raise TypeError("Only seconds or frames may be given, not both.")

        start = self.clock.t
        for f in count(1):
            await self.next_frame()
            now = self.clock.t - start
            if seconds is not None and now >= seconds:
                yield seconds
                return

            yield now
            if f == frames:
                break

    async def frames_dt(self, *, seconds=None, frames=None):
        """Iterate over multiple frames, yielding the time per frame."""
        last_t = 0
        async for t in self.frames(seconds=seconds, frames=frames):
            yield t - last_t
            last_t = t

    async def interpolate(self, start, end, duration=1.0, tween='linear'):
        """Iterate over values between start and end, over the given duration.

        The values of 'tween' are as for animate().

        For example,

            async for pos in clock.coro.tween(ship.pos, target_pos, 1.0):
                space_ship.pos = pos

        """
        from . import animation
        func = animation.TWEEN_FUNCTIONS[tween]

        async for t in self.frames(seconds=duration):
            if t >= duration:
                yield end
                return
            frac = func(t / duration)
            yield animation.tween_attr(frac, start, end)

    def run(self, coro):
        """Schedule a coroutine."""
        assert isinstance(coro, types.CoroutineType)
        task = Task(self.clock, coro)
        return task


class Task:
    def __init__(self, clock, coro):
        self.clock = clock
        self.coro = coro
        self.result = None
        self._step()

    def _step(self, await_value=None):
        """Step this task, passing the given value that was awaited.

        For time events, await_value will be a time delta in seconds; for
        condition events, it will be some other value.

        """
        clock = self.clock
        clock.unschedule(self._step)

        if self.coro is None:
            return

        try:
            res = self.coro.send(await_value)
        except StopIteration as stop:
            if stop.args:
                self.result = stop.args[0]
            return

        if not isinstance(res, Future):
            raise TypeError(
                f"Unable to await {res!r} with "
                "clock.coro.run(). wasabi2d coroutines are not "
                "compatible with asyncio."
            )

        val = res.val
        if isinstance(val, WaitDelay):
            clock.schedule(self._step, val.seconds, strong=True)
        elif val is WaitTick:
            clock.call_soon(self._step)
        elif isinstance(val, WaitEvent):
            # Bit ugly
            val.when_ready(
                lambda: clock.call_soon(
                    lambda dt: self._step(val._result)
                )
            )
        else:
            raise TypeError("Unexpected value")

    def cancel(self):
        """Cancel the task."""
        try:
            self.coro.throw(Coroutines.Cancelled)
        except (StopIteration, Coroutines.Cancelled):
            # Coroutine halted successfully. If not it may be awaiting
            # something during exception handling, and we should not unschedule
            # it.
            self.clock.unschedule(self._step)
            self.coro = None


class Clock:
    """A clock used for event scheduling.

    When tick() is called, all events scheduled for before now will be called
    in order.

    tick() would typically be called from the game loop for the default clock.

    Additional clocks could be created - for example, a game clock that could
    be suspended in pause screens. Your code must take care of calling tick()
    or not. You could also run the clock at a different rate if desired, by
    scaling dt before passing it to tick().

    """
    def __init__(self):
        self.t = 0
        self.paused = False
        self.fired = False
        self.events = []
        self._each_tick = []
        self._next_tick = []
        self.coro = Coroutines(self)

    def clear(self):
        """Remove all handlers from this clock."""
        self.events.clear()
        self._each_tick.clear()

    def schedule(self, callback, delay, *, strong=False):
        """Schedule callback to be called once, at `delay` seconds from now.

        :param callback: A parameterless callable to be called.
        :param delay: The delay before the call (in clock time / seconds).

        """
        heapq.heappush(
            self.events,
            Event(self.t + delay, callback, strong, None)
        )

    def schedule_unique(self, callback, delay, *, strong=False):
        """Schedule callback to be called once, at `delay` seconds from now.

        If it was already scheduled, postpone its firing.

        :param callback: A parameterless callable to be called.
        :param delay: The delay before the call (in clock time / seconds).

        """
        self.unschedule(callback)
        self.schedule(callback, delay, strong=strong)

    def schedule_interval(self, callback, delay, *, strong=False):
        """Schedule callback to be called every `delay` seconds.

        The first occurrence will be after `delay` seconds.

        :param callback: A parameterless callable to be called.
        :param delay: The interval in seconds.

        """
        heapq.heappush(
            self.events,
            Event(self.t + delay, callback, strong, delay)
        )

    def unschedule(self, callback):
        """Unschedule the given callback.

        If scheduled multiple times all instances will be unscheduled.

        """
        self.events = [
            e for e in self.events
            if e.callback != callback and e.callback is not None
        ]
        heapq.heapify(self.events)
        self._each_tick = [e for e in self._each_tick if e() != callback]

    def call_soon(self, callback):
        """Schedule a function to be called on the next tick.

        The function will receive a parameter `dt` indicating the time that
        has passed.

        The callback will always be strongly referenced.

        """
        self._next_tick.append(lambda: callback)

    def each_tick(self, callback, strong=False):
        """Schedule a callback to be called every tick.

        Unlike the standard scheduler functions, the callable is passed the
        elapsed clock time since the last call (the same value passed to tick).

        """
        self._each_tick.append(
            (lambda: callback) if strong else mkref(callback)
        )

    def _fire_each_tick(self, dt):
        dead = [
            None,  # None means a weak ref has expired, always remove
        ]
        to_fire = chain(self._next_tick, self._each_tick)
        self._next_tick = []
        for r in to_fire:
            cb = r()
            if cb is not None:
                self.fired = True
                try:
                    cb(dt)
                except Exception:
                    import traceback
                    traceback.print_exc()
                    dead.append(cb)
        self._each_tick = [e for e in self._each_tick if e() not in dead]

    def tick(self, dt: float) -> bool:
        """Update the clock time and fire all scheduled events.

        :param dt: The elapsed time in seconds.
        :return bool: Return True if any callback was triggered.

        """
        if self.paused:
            return False
        self.fired = False
        self.dt = dt = float(dt)
        self.t += dt
        self._fire_each_tick(dt)
        while self.events and self.events[0].time <= self.t:
            ev = heapq.heappop(self.events)
            cb = ev.callback
            if not cb:
                continue

            if ev.repeat is not None:
                self.schedule_interval(cb, ev.repeat)

            self.fired = True
            try:
                cb()
            except Exception:
                import traceback
                traceback.print_exc()
                self.unschedule(cb)
        return self.fired


# One instance of a clock is available by default, to simplify the API
default_clock = clock = Clock()
tick = clock.tick
schedule = clock.schedule
schedule_interval = clock.schedule_interval
schedule_unique = clock.schedule_unique
unschedule = clock.unschedule
each_tick = clock.each_tick
call_soon = clock.call_soon
coro = clock.coro



"""Base functionality for post-processing effects."""
import numpy as np
import moderngl

from ..shaders import shadermgr


class PostprocessPass:
    """A full-screen postprocessing effect pass.

    This will render a full-screen quad using a single fragment shader.

    """
    # TODO: let the shadermanager itself manage these, so that they
    # are shared between layers etc

    QUAD_INDICES = np.array([0, 1, 2, 0, 2, 3], dtype='i4')
    QUAD_VERTS_UVS = np.array([
        [-1, -1, 0, 0],
        [1, -1, 1, 0],
        [1, 1, 1, 1],
        [-1, 1, 0, 1],
    ], dtype='f4')

    def __init__(
            self,
            ctx: moderngl.Context,
            shader_name: str=None,
            text: str=None,
            send_uvs: bool = True):
        self.ctx = ctx
        if shader_name:
            self.prog = shadermgr(ctx).load(
                'postprocess/postprocess',
                shader_name
            )
        else:
            sm = shadermgr(ctx)
            self.prog = sm.get(
                sm._read('postprocess/postprocess.vert'),
                text
            )

        self.indices = ctx.buffer(self.QUAD_INDICES)
        self.vs_uvs = ctx.buffer(self.QUAD_VERTS_UVS)

        if send_uvs:
            attribs = (self.vs_uvs, '2f4 2f4', 'in_vert', 'in_uv')
        else:
            attribs = (self.vs_uvs, '2f4 8x', 'in_vert')
        self.vao = ctx.vertex_array(
            self.prog,
            [attribs],
            index_buffer=self.indices
        )

    def __del__(self):
        self.vao.release()
        self.indices.release()
        self.vs_uvs.release()

    def set_region(self, xfrac=1, yfrac=1):
        """Set the pass to write to a subregion of the viewport."""
        coords = self.QUAD_VERTS_UVS.copy()
        coords[:, 0] *= xfrac
        coords[:, 0] -= (1.0 - xfrac)
        coords[:, 1] *= yfrac
        coords[:, 1] -= (1.0 - yfrac)
        coords[:, 2] *= xfrac
        coords[:, 3] *= yfrac
        self.vs_uvs.write(coords)

    def render(self, **uniforms):
        """Assign the given uniforms and then render."""
        texnum = 0
        for k, v in uniforms.items():
            if isinstance(v, moderngl.Framebuffer):
                v = v.color_attachments[0]

            if isinstance(v, moderngl.Texture):
                v.use(texnum)
                v = texnum
                texnum += 1

            if isinstance(v, np.ndarray):
                self.prog[k].write(v)
            else:
                self.prog[k].value = v
        self.vao.render(moderngl.TRIANGLES, 6)



"""Allocator for index buffers.

Without indirect rendering (unavailable on Mac OS X) we have to rely on packed
index buffers with primitive restarts where necessary. This requires rebuilding
the index buffer itself when anything changes, which is likely to be more
expensive than just rebuilding the indirect buffer.

Users must insert their own primitive restart commands.

These buffers support sorting of the allocations, which can be used to
implement depth-sorted or y-sorted layers.

"""
from typing import Any, Dict, Tuple, Iterable, Sequence, Mapping
import heapq
from functools import total_ordering

import numpy as np
import moderngl as mgl
from sortedcontainers import SortedDict


class IndexBuffer:
    """Abstraction over index lists.

    We use integer identifiers to track the allocations, for encapsulation.
    These are generated anyway as a way to preserve insertion order in the
    absence of any other sort key.

    """

    def __init__(self, ctx: mgl.Context):
        self.buffer = None
        self.ctx = ctx

        # Allocation id to sort key
        self.id_lookup: Dict[int, Tuple[Any, int]] = {}

        # Sorted list of sort keys to index arrays
        self.allocations: Mapping[Tuple[Any, int], np.ndarray] = SortedDict()

        # Track whether we have updates
        self.dirty: bool = True

        # We allocate identifiers for each allocation. These are sequential
        # and form part of the sort key; this ensures that insertion order
        # can be preserved.
        self.next_id: int = 1  # start at 1 so we can test bool(id)

    def insert(self, indexes: np.ndarray, sort: Any = None) -> int:
        """Add indexes to the buffer."""
        assert indexes.dtype == np.uint32, \
            f"incorrect dtype {indexes.dtype!r}, expected uint32"
        id = self.next_id
        self.next_id += 1

        indexes.flags.writeable = False
        k = sort, id
        self.id_lookup[id] = k
        self.allocations[k] = indexes
        self.dirty = True
        return id

    def remove(self, id: int):
        """Remove an index range."""
        k = self.id_lookup.pop(id)
        del self.allocations[k]
        self.dirty = True

    def clear(self):
        """Clear all allocations."""
        self.allocations.clear()
        self.id_lookup.clear()
        self.next_id = 0
        self.dirty = True

    def __contains__(self, id: int) -> bool:
        """Return True if the given id is allocated."""
        return id in self.id_lookup

    def __bool__(self) -> bool:
        return bool(self.id_lookup)

    def set_indexes(self, id: int, indexes: np.ndarray):
        """Replace the indexes for an allocation."""
        assert indexes.dtype == np.uint32, \
            f"incorrect dtype {indexes.dtype!r}, expected uint32"
        indexes.flags.writeable = False
        k = self.id_lookup[id]
        self.allocations[k] = indexes
        self.dirty = True

    def set_sort(self, id: int, sort: Any):
        """Set the sort key for an allocation."""
        k = self.id_lookup[id]
        indexes = self.allocations.pop(k)
        k = sort, id
        self.id_lookup[id] = k
        self.allocations[k] = indexes
        self.dirty = True

    def update(self, id: int, indexes: np.ndarray, sort: Any = None):
        """Update sort and indexes for an allocation."""
        assert indexes.dtype == np.uint32, \
            f"incorrect dtype {indexes.dtype!r}, expected uint32"
        indexes.flags.writeable = False
        k = self.id_lookup.pop(id)
        del self.allocations[k]

        k = sort, id
        self.id_lookup[id] = k
        self.allocations[k] = indexes
        self.dirty = True

    def as_array(self) -> np.ndarray:
        """Flatten the allocations to a numpy array."""
        # TODO: maybe track the total length of the allocations and keep a
        # memoized array here. This would let us hstack into an existing array
        # if our allocations have simply changed sort key.
        return np.hstack(self.allocations.values())

    def get_buffer(self) -> mgl.Buffer:
        """Get the index buffer."""
        if self.dirty:
            if self.buffer:
                # TODO: use moderngl orphan with resize
                self.buffer.release()
            self.buffer = self.ctx.buffer(self.as_array())
        return self.buffer

    def release(self):
        """Release the buffer, if allocated."""
        if self.buffer:
            self.buffer.release()
            self.buffer = None
            self.dirty = True

    __del__ = release


@total_ordering
class BufIter:
    """An iterator for the merge."""
    __slots__ = ('it', 'peeked', 'idxs', 'idxpos')

    def __init__(self, buffer: IndexBuffer):
        self.it = iter(buffer.allocations.items())
        self.peeked, self.idxs = next(self.it)
        self.idxpos = 0

    def next(self):
        item = self.peeked, self.idxs
        self.idxpos += len(self.idxs)
        self.peeked, self.idxs = next(self.it, None)
        return item

    def __lt__(self, other):
        return self.peeked < other.peeked

    def __eq__(self, other):
        return self.peeked == other.peeked


def bufiter(buf: IndexBuffer) -> Iterable[Tuple[Any, IndexBuffer, int, int]]:
    """Iterate over allocations in a buffer."""
    pos = 0
    for (sort, id), indexes in buf.allocations.items():
        nextpos = pos + len(indexes)
        yield sort, buf, pos, nextpos
        pos = nextpos


def merge_seq(
    buffers: Sequence[IndexBuffer]
) -> Iterable[Tuple[IndexBuffer, int, int]]:
    """Merge a set of index buffers sorted by a common sort key.

    Yield contiguous groups of indexes.

    This implementation is O(n) in the number of allocations. We could do
    better - closer to O(m log n) for n allocations and m switches - if we
    had cached offsets for each index range. Then we could bisect into a
    buffer and read out the offset. This would still be an optimisation if we
    can avoid rebuilding that cache every frame. This is true if primitives
    do not move.

    """
    bufiters = map(bufiter, buffers)
    merged_allocs = heapq.merge(*bufiters, key=lambda i: (i[0], id(i[1])))
    try:
        _, lastbuf, laststart, lastend = next(merged_allocs)
    except StopIteration:
        return
    for _, buf, start, end in merged_allocs:
        if buf is not lastbuf:
            yield lastbuf, laststart, lastend
            lastbuf = buf
            laststart = start
            lastend = end
        else:
            lastend = end
    yield lastbuf, laststart, lastend



import random
import wasabi2d as w2d
import colorsys
from wasabi2d import clock, Vector2, animate, keyboard


scene = w2d.Scene(title="Run!")
scene.background = 'white'
target = Vector2(scene.width / 2, scene.height / 2)
scene.layers[0].set_effect('dropshadow', opacity=1, radius=1)


async def spawn_baddies():
    while True:
        clock.coro.run(enemy())
        await clock.coro.sleep(3)


async def enemy():
    color = colorsys.hsv_to_rgb(random.random(), 1, 1)
    pos = Vector2(
        random.uniform(50, scene.width - 50),
        random.uniform(50, scene.height - 50)
    )
    e = scene.layers[0].add_circle(
        radius=10,
        color=color,
        pos=pos,
    )
    e.scale = 0.1
    await animate(
        e,
        duration=0.3,
        scale=1,
    )
    async for dt in clock.coro.frames_dt():
        to_target = target - pos
        if to_target.magnitude() < e.radius:
            break
        pos += to_target.normalize() * 100 * dt
        e.pos = pos

    await animate(
        e,
        duration=0.5,
        scale=4,
        tween='accelerate',
        color=(*color, 0)
    )

    e.delete()


@w2d.event
def on_mouse_move(pos):
    global target
    target = Vector2(*pos)


clock.coro.run(spawn_baddies())
w2d.run()



from pathlib import Path
from setuptools import setup, find_packages

root = Path(__file__).parent
LONG_DESCRIPTION = root.joinpath('README.md').read_text(encoding='utf8')


def package_data(dirname):
    w2d = root / 'wasabi2d'
    data_dir = w2d / dirname
    yield f'{dirname}/*'
    for path in data_dir.glob('**/'):
        yield f'{path.relative_to(w2d)}/*'


setup(
    name='wasabi2d',
    version='1.4.0',
    description="A convenient 2D OpenGL games framework",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/lordmauve/wasabi2d',
    author='Daniel Pope',
    author_email='mauve@mauveweb.co.uk',
    packages=find_packages(include='wasabi2d*'),
    package_data={'wasabi2d': [
        *package_data('data'),
        *package_data('glsl'),
    ]},
    install_requires=open('requirements.txt').read().splitlines(),
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Education',
        'Topic :: Games/Entertainment',
    ],
    test_suite='tests'
)



"""Pixellate by averaging pixel values."""
from dataclasses import dataclass

import moderngl

from ..shaders import bind_framebuffer, blend_func
from .base import PostprocessPass


@dataclass
class Pixellate:
    """A pixellation effect."""

    ctx: moderngl.Context
    pxsize: int = 10
    antialias: float = 1.0

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera

        self._average = PostprocessPass(
            self.ctx,
            'postprocess/pixellate_average',
        )
        self._fill = PostprocessPass(
            self.ctx,
            'postprocess/pixellate_copy',
        )

    def draw(self, draw_layer):
        # Fraction to reduce by each pass
        frac = 1 / self.pxsize

        # By turning off the averaging we can remove the antialiasing
        epxsize = round((self.pxsize - 1) * self.antialias) + 1

        with self.camera.temporary_fbs(2, 'f2') as (fb1, fb2):
            with bind_framebuffer(self.ctx, fb1, clear=True):
                draw_layer()

                with blend_func(self.ctx, moderngl.ONE, moderngl.ZERO):
                    with bind_framebuffer(self.ctx, fb2, clear=True):
                        # Pass 1: downsample by frac in the y direction
                        self._average.set_region(1, frac)
                        self._average.render(
                            image=fb1,
                            blur_direction=(0, 1),
                            pxsize=epxsize,
                            uvscale=(1, self.pxsize),
                        )

                    # Pass 2: downsample by frac in the x direction
                    self._average.set_region(frac, frac)
                    self._average.render(
                        image=fb2,
                        blur_direction=(1, 0),
                        pxsize=epxsize,
                        uvscale=(self.pxsize, 1)
                    )

            # Pass 3: scale the downsampled image to the screen
            fb1_tex = fb1.color_attachments[0]
            lin = fb1_tex.filter
            fb1_tex.filter = moderngl.NEAREST, moderngl.NEAREST
            self._fill.render(image=fb1, pxsize=self.pxsize)
            fb1_tex.filter = lin



"""Common descriptors to simplify setting attributes on objects."""


class CallbackProp:
    """Descriptor for a property that calls a callback when set.

    The callback is passed the instance that the attribute was set on.

    """
    __slots__ = ('callback', 'name')

    def __init__(self, callback):
        self.callback = callback
        self.name = None

    def __set_name__(self, cls, name):
        self.name = f'_{name}'

    def __get__(self, inst, cls=None):
        return getattr(inst, self.name)

    def __set__(self, inst, value):
        setattr(inst, self.name, value)
        self.callback(inst)



"""Example of the light bloom effect."""
from wasabi2d import run, Scene, event, Vector2
from wasabi2d.actor import Actor


scene = Scene()
scene.background = 'red'

logo = Actor(
    scene.layers[0].add_sprite(
        'wasabi2d',
        pos=(scene.width / 2, scene.height / 2),
    )
)
logo.v = Vector2(1, -1)
scene.layers[0].set_effect(
    'trails',
    fade=0.7
)


@event
def update(dt):
    logo.pos += logo.v
    if logo.top < 0 or logo.bottom >= scene.height:
        logo.v.y *= -1
    if logo.left < 0 or logo.right >= scene.width:
        logo.v.x *= -1


run()
